from __future__ import annotations

import json
from itertools import chain
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, cast

import pydash.collections
from fideslang.validation import FidesKey
from loguru import logger
from sqlalchemy.orm import Session

from fides.api.common_exceptions import (
    TraversalError,
    UnreachableEdgesError,
    UnreachableNodesError,
)
from fides.api.graph.config import (
    ROOT_COLLECTION_ADDRESS,
    TERMINATOR_ADDRESS,
    Collection,
    CollectionAddress,
    FieldAddress,
    FieldPath,
    GraphDataset,
)
from fides.api.graph.execution import ExecutionNode
from fides.api.graph.graph import DatasetGraph, Edge, Node
from fides.api.graph.node_filters import (
    CustomRequestFieldFilter,
    NodeFilter,
    OptionalIdentityFilter,
    PolicyDataCategoryFilter,
)
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import (
    PrivacyRequest,
    RequestTask,
    TraversalDetails,
)
from fides.api.schemas.policy import ActionType
from fides.api.util.collection_util import Row, append, partition
from fides.api.util.logger_context_utils import Contextualizable, LoggerContextKeys
from fides.api.util.matching_queue import MatchingQueue

ARTIFICIAL_NODES: List[CollectionAddress] = [
    ROOT_COLLECTION_ADDRESS,
    TERMINATOR_ADDRESS,
]

Datastore = Dict[CollectionAddress, List[Row]]
"""A type expressing retrieved rows of data from a specified collection"""


def artificial_traversal_node(address: CollectionAddress) -> TraversalNode:
    """generate an 'artificial' traversal_node pointing to the given address. This is used to
    generate artificial root and termination nodes that correspond to just an address, but
    have no actual corresponding collection dataset"""
    ds: Collection = Collection(name=address.collection, fields=[])
    node = Node(
        GraphDataset(
            name=address.dataset, collections=[ds], connection_key="__IGNORE__"
        ),
        ds,
    )
    return TraversalNode(node)


class BaseTraversal:
    """Handling for a single reified traversal of a graph based on input (seed) data."""

    def extract_seed_field_addresses(
        self,
    ) -> Dict[FieldAddress, str]:
        """Find seeded values in the graph based on the input seed and return
        a list of the populated seed nodes"""

        return {
            identity_address: seed_key
            for identity_address, seed_key in self.graph.identity_keys.items()
            if seed_key in self.seed_data
        }

    def __init__(
        self,
        graph: DatasetGraph,
        data: Dict[str, Any],
        *,
        policy: Optional[Policy] = None,
        node_filters: Optional[List[NodeFilter]] = None,
    ):
        self.graph = graph
        self.seed_data = data
        self.policy = policy

        # Node filters are used to allow unreachable nodes to be ignored
        # if they don't apply to the given scenario
        self.node_filters = node_filters or []

        self.traversal_node_dict = {k: TraversalNode(v) for k, v in graph.nodes.items()}
        self.edges: Set[Edge] = graph.edges.copy()
        self.root_node = artificial_traversal_node(ROOT_COLLECTION_ADDRESS)
        for (
            start_field_address,
            seed_key,
        ) in self.extract_seed_field_addresses().items():
            self.edges.add(
                Edge(
                    FieldAddress(
                        ROOT_COLLECTION_ADDRESS.dataset,
                        ROOT_COLLECTION_ADDRESS.collection,
                        seed_key,
                    ),
                    start_field_address,
                )
            )

        # Ensure manual_task collections execute right after ROOT
        from fides.api.task.manual.manual_task_address import ManualTaskAddress

        for addr in self.traversal_node_dict.keys():
            if ManualTaskAddress.is_manual_task_address(addr):
                # Add a simple synthetic edge ROOT.id -> manual_data.id
                self.edges.add(
                    Edge(
                        FieldAddress(
                            ROOT_COLLECTION_ADDRESS.dataset,
                            ROOT_COLLECTION_ADDRESS.collection,
                            "id",
                        ),
                        addr.field_address(FieldPath("id")),
                    )
                )

        self._verify_traversal()

    def _verify_traversal(self) -> None:
        """Verify that a valid traversal exists. This method simply assembles a traversal
        and raises an error on any traversal failure conditions."""
        self.traverse(
            {self.root_node.address: [self.seed_data]},
            lambda n, m: logger.info("Traverse {}", n.address),
        )

    def traversal_map(
        self,
    ) -> Tuple[Dict[str, List[Dict[str, Any]]], List[CollectionAddress]]:
        """Generate a descriptive map of the traversal generated.

        For each traversal_node N, list
         { N.address:
             {"from": {traversal_node address:
                       [field in "from", field in N for all edges from "from" to N]}}
             {"to": {traversal_node address,
                       [field in N, field in "to"" for all edges from N to "to"]}}

        """

        def traversal_dict_fn(
            tn: TraversalNode, data: Dict[CollectionAddress, Dict[str, Any]]
        ) -> None:
            data[tn.address] = tn.debug()

        db = {ROOT_COLLECTION_ADDRESS: [self.seed_data]}
        traversal_ends = self.traverse(db, traversal_dict_fn)

        return {str(k): v for k, v in db.items()}, traversal_ends

    def should_exclude_node(self, node: TraversalNode) -> bool:
        """Returns True if any filter indicates the node should be excluded"""
        return any(filter.exclude_node(node) for filter in self.node_filters)

    def traverse(  # pylint: disable=R0914
        self,
        environment: Dict[CollectionAddress, Any],
        node_run_fn: Callable[[TraversalNode, Dict[CollectionAddress, Any]], None],
    ) -> List[CollectionAddress]:
        """Traverse and call run() on each traversal_node in turn. T represents an environment that
        can provide or collect values as each traversal_node is run.

        Returns a list of termination traversal_node addresses so that we can take action on completed
        traversal.

        We define the root traversal_node as a traversal_node whose children are any nodes that have identity (seed)
        data.
        We start with
        - a queue holding only the root traversal_node.
        - a (copied) set of all of the edges in the graph.

        - Pop the first eligible traversal_node from the queue.
        - call the passed in callable node_run_function on this traversal_node. Mark this traversal_node as "finished"
        - Delete all edges from any finished nodes to this traversal_node.
        - put all of this nodes children in the queue.

        Some nodes have conditions, like "don't run me until after traversal_node X". When we pop a value from
        the queue, we are taking this into account. If the queue contains nodes, but none of them are
        eligible (e.g. Node A can't run until after B, and traversal_node B that can't run until after A)
        raise a TraversalError.

        We also raise a TraversalError if the queue is empty but some nodes have not been visited. In
        that case they are unreachable.

        """
        if environment:
            logger.info(
                "Starting traversal",
            )
        remaining_node_keys: Set[CollectionAddress] = set(
            self.traversal_node_dict.keys()
        )
        finished_nodes: dict[CollectionAddress, TraversalNode] = {}
        running_node_queue: MatchingQueue[TraversalNode] = MatchingQueue(self.root_node)

        remaining_edges: Set[Edge] = self.edges.copy()
        while not running_node_queue.is_empty():
            # this is to support the "run traversal_node A AFTER traversal_node B functionality:"
            n = running_node_queue.pop_first_match(
                lambda x: x.can_run_given(remaining_node_keys)
            )

            if n:
                node_run_fn(n, environment)
                # delete all edges between the traversal_node that's just run and any completed nodes
                for finished_node_address, finished_node in finished_nodes.items():
                    completed_edges: Set[Edge] = Edge.delete_edges(
                        remaining_edges,
                        finished_node_address,
                        cast(TraversalNode, n).address,  # type: ignore[redundant-cast]
                    )

                    def edge_ends_with_collection(_edge: Edge) -> bool:
                        # append edges that end in this traversal_node
                        return _edge.ends_with_collection(
                            cast(TraversalNode, n).address  # type: ignore[redundant-cast]
                        )

                    for edge in filter(edge_ends_with_collection, completed_edges):
                        # note, this will not work for self-reference
                        finished_node.add_child(n, edge)
                # next edges = take all edges including n that are _not_ in edges_from_completed_nodes
                # in the form (field_address_this, field_address_foreign)

                edges_to_children = pydash.collections.filter_(
                    [
                        e.split_by_address(cast(TraversalNode, n).address)  # type: ignore[redundant-cast]
                        for e in remaining_edges
                        if e.contains(n.address)
                    ]
                )
                if not edges_to_children:
                    n.is_terminal_node = True

                # child traversal_node addresses are the address portion of the above
                child_node_addresses = {
                    a[1].collection_address() for a in edges_to_children if a
                }
                for nxt_address in child_node_addresses:
                    # only add the next traversal_node to the queue if it is not already there (no duplicates)
                    running_node_queue.push_if_new(
                        self.traversal_node_dict[nxt_address]
                    )
                finished_nodes[n.address] = n
                remaining_node_keys.difference_update({n.address})
            else:
                # traversal traversal_node dict diff finished nodes
                logger.error(
                    "Node could not be reached given specified ordering [{}]",
                    ", ".join([str(tn.address) for tn in running_node_queue.data]),
                )
                raise TraversalError(
                    f"""Node could not be reached given the specified ordering:
                    [{', '.join([str(tn.address) for tn in running_node_queue.data])}]""",
                )

        remaining_node_keys = {
            key
            for key in remaining_node_keys
            if not self.should_exclude_node(self.traversal_node_dict[key])
        }

        # error if there are nodes that have not been visited
        if remaining_node_keys:
            logger.error(
                "Some nodes were not reachable: {}",
                ", ".join([str(x) for x in remaining_node_keys]),
            )
            raise UnreachableNodesError(
                f"Some nodes were not reachable: {', '.join([str(x) for x in remaining_node_keys])}",
                [key.value for key in remaining_node_keys],
            )

        # filter out remaining_edges if the nodes they link are allowed to remain unreachable
        remaining_edges = {
            edge
            for edge in remaining_edges
            if edge.f1.collection_address() in remaining_node_keys
            and edge.f2.collection_address() in remaining_node_keys
        }

        # error if there are edges that have not been visited
        if remaining_edges:
            logger.error(
                "Some edges were not reachable: {}",
                ", ".join([str(x) for x in remaining_edges]),
            )
            raise UnreachableEdgesError(
                f"Some edges were not reachable: {', '.join([str(x) for x in remaining_edges])}",
                [f"{edge}" for edge in remaining_edges],
            )

        end_nodes = [
            tn.address for tn in finished_nodes.values() if tn.is_terminal_node
        ]
        if environment:
            logger.debug("Found {} end nodes: {}", len(end_nodes), end_nodes)
        return end_nodes

    @property
    def skipped_nodes(self) -> Dict[str, str]:
        """
        Returns a dictionary of node addresses to skip reasons from all node filters.
        """
        result = {}
        for node_filter in self.node_filters:
            result.update(node_filter.skipped_nodes)
        return result


class Traversal(BaseTraversal):
    """Handling for a single reified traversal of a graph based on input (seed) data."""

    def __init__(
        self,
        graph: DatasetGraph,
        data: Dict[str, Any],
        *,
        policy: Optional[Policy] = None,
        node_filters: Optional[List[NodeFilter]] = None,
    ):
        filters = node_filters or []

        filters.append(CustomRequestFieldFilter())
        filters.append(OptionalIdentityFilter(graph, data))
        if policy:
            filters.append(PolicyDataCategoryFilter(policy))

        super().__init__(graph, data, policy=policy, node_filters=filters)


def log_traversal_error_and_update_privacy_request(
    privacy_request: PrivacyRequest, session: Session, err: TraversalError
) -> None:
    """
    Logs the provided traversal error with the privacy request id, creates the corresponding
    ExecutionLog instances, and marks the privacy request as errored.

    If the error is a generic TraversalError, a generic error execution log is created.
    If the error is an UnreachableNodesError or UnreachableEdgesError, an execution log is created
    for each node / edge on the "errors" list of the exception.
    """
    logger.error(
        "TraversalError encountered for privacy request. Error: {}",
        err,
    )

    # For generic TraversalErrors, we log a generic error execution log
    if not isinstance(err, UnreachableNodesError) and not isinstance(
        err, UnreachableEdgesError
    ):
        privacy_request.add_error_execution_log(
            session,
            connection_key=None,
            dataset_name="Dataset traversal",
            collection_name=None,
            message=str(err),
            action_type=ActionType.access,
        )

    # For specific ones, we iterate over each unreachable node key in the list
    for error in err.errors:
        dataset, collection = (
            error.split(":")
            if isinstance(
                err, UnreachableNodesError
            )  # For unreachable nodes, we can get the dataset and collection from the node
            else (None, None)  # But not for edges
        )
        message = f"{'Node' if isinstance(err, UnreachableNodesError) else 'Edge'} {error} is not reachable"
        privacy_request.add_error_execution_log(
            session,
            connection_key=None,
            dataset_name="Dataset traversal",
            collection_name=f"{dataset}.{collection}",
            message=message,
            action_type=ActionType.access,
        )
    privacy_request.error_processing(session)


class TraversalNode(Contextualizable):
    """Traversal_node type. This type is used for building the graph, not for executing the graph."""

    def __init__(self, node: Node):
        self.node = node
        self.address = node.address
        # address to a child that holds both a traversal_node and all possible paths
        # from this traversal_node to it: {address -> [(traversal_node, from_field, to_field)}
        self.children: Dict[
            CollectionAddress, List[Tuple[TraversalNode, FieldPath, FieldPath]]
        ] = {}
        self.parents: Dict[
            CollectionAddress, List[Tuple[TraversalNode, FieldPath, FieldPath]]
        ] = {}
        self.is_terminal_node = False

    def add_child(self, child_node: TraversalNode, edge: Edge) -> None:
        """Add other as a child to this traversal_node along the provided edge."""
        addresses = edge.split_by_address(self.address)  # (traversal_node -> other)
        if addresses:
            self_field_address, other_field_address = addresses
            append(
                self.children,
                other_field_address.collection_address(),
                (
                    child_node,
                    self_field_address.field_path,
                    other_field_address.field_path,
                ),
            )
            append(
                child_node.parents,
                self_field_address.collection_address(),
                (self, self_field_address.field_path, other_field_address.field_path),
            )

    def incoming_edges(self) -> Set[Edge]:
        """Return the incoming edges to this traversal_node,in (other.address -> self.address) order."""
        return {
            Edge(
                p_collection_address.field_address(parent_field_path),
                self.address.field_address(self_field_path),
            )
            for p_collection_address, tuples in self.parents.items()
            for _, parent_field_path, self_field_path in tuples
        }

    def outgoing_edges(self) -> Set[Edge]:
        """Return the outgoing edges to this traversal_node,in (self.address -> other.address) order."""
        return {
            Edge(
                self.address.field_address(self_field_path),
                c_collection_address.field_address(child_field_path),
            )
            for c_collection_address, tuples in self.children.items()
            for _, self_field_path, child_field_path in tuples
        }

    def incoming_edges_by_collection(self) -> Dict[CollectionAddress, List[Edge]]:
        return partition(self.incoming_edges(), lambda e: e.f1.collection_address())

    def input_keys(self) -> List[CollectionAddress]:
        """Returns the inputs to the current node that are data dependencies
        This is copied and saved to the RequestTask and used to maintain a consistent order
        for passing in data for an access task
        """
        return sorted(self.incoming_edges_by_collection().keys())

    def can_run_given(self, remaining_node_keys: Set[CollectionAddress]) -> bool:
        """True if finished_node_keys covers all the nodes that this traversal_node is waiting for.  If
        all nodes this traversal_node is waiting for have finished, it's ok for this traversal_node to run.

        NOTE: "After" functionality may not work as expected.
        """
        if self.node.collection.after.intersection(
            remaining_node_keys
        ) or self.node.dataset.after.intersection(
            {k.dataset for k in remaining_node_keys}
        ):
            return False
        return True

    def is_root_node(self) -> bool:
        """This traversal_node is the defined traversal start"""
        return self.address == ROOT_COLLECTION_ADDRESS

    def debug(self) -> Dict[str, Any]:
        """Generate debug descriptive output of the traversal."""
        _from: Dict[str, Set[str]] = {}

        for edge in self.incoming_edges():
            foreign = edge.f1
            local = edge.f2
            # if we have a bidirectional edge, make sure that we correctly label
            # which edge is foreign.
            if foreign.collection_address() == self.address:
                foreign, local = local, foreign
            key = str(foreign.collection_address())
            val = f"{foreign.field_path.string_path} -> {local.field_path.string_path}"
            if key in _from:
                _from[key].add(val)
            else:
                _from[key] = {val}
        to: Dict[str, Set[str]] = {}
        for k, v in self.children.items():
            to[str(k)] = {f"{f[1].string_path} -> {f[2].string_path}" for f in v}
        return {
            "from": {k: set(v) for k, v in _from.items()},
            "to": {k: set(v) for k, v in to.items()},
        }

    def get_log_context(self) -> Dict[LoggerContextKeys, Any]:
        return {LoggerContextKeys.collection: self.node.collection.name}

    def format_traversal_details_for_save(self) -> Dict:
        """Convert key traversal details from the TraversalNode for save on the RequestTask.

        The RequestTask will be retrieved from the database and the traversal details
        used to build the ExecutionNode for DSR 3.0.
        """

        connection_key: FidesKey = self.node.dataset.connection_key

        return TraversalDetails(
            dataset_connection_key=connection_key,
            incoming_edges=[
                [edge.f1.value, edge.f2.value] for edge in self.incoming_edges()
            ],
            outgoing_edges=[
                [edge.f1.value, edge.f2.value] for edge in self.outgoing_edges()
            ],
            input_keys=[tn.value for tn in self.input_keys()],
        ).model_dump(mode="json")

    def to_mock_request_task(self) -> RequestTask:
        """Converts a portion of the TraversalNode into a RequestTask - used in building
        dry run queries or for supporting Deprecated DSR 2.0. Request Tasks were introduced in DSR 3.0
        """
        collection_data = json.loads(
            # Serializes with duck-typing behavior, no longer the default in Pydantic v2
            # Needed for serializing nested collection fields
            self.node.collection.model_dump_json(serialize_as_any=True)
        )
        return RequestTask(  # Mock a RequestTask object in memory
            collection_address=self.node.address.value,
            dataset_name=self.node.address.dataset,
            collection_name=self.node.address.collection,
            collection=collection_data,
            traversal_details=self.format_traversal_details_for_save(),
        )

    def to_mock_execution_node(self) -> ExecutionNode:
        """Converts a TraversalNode into an ExecutionNode - used for supporting DSR 2.0, to convert
        Traversal Nodes into the Execution Node format which is needed for executing the graph in
        DSR 3.0

        DSR 3.0 on the other hand, creates ExecutionNodes from data on the RequestTask.
        """
        request_task: RequestTask = self.to_mock_request_task()
        return ExecutionNode(request_task)

    def get_data_categories(self) -> Set[str]:
        """
        Returns a set of unique data categories across the collection and child fields.
        """
        collection = self.node.collection

        return set(
            chain(
                collection.data_categories,
                *[
                    field.data_categories or []
                    for field in collection.field_dict.values()
                ],
            )
        )



from enum import Enum
from typing import Optional

from pydantic import Field, SerializeAsAny
from pydantic_settings import SettingsConfigDict

from fides.api.custom_types import AnyHttpUrlStringRemovesSlash

from .fides_settings import FidesSettings


class ErrorNotificationMode(str, Enum):
    CONSOLE_ONLY = "console_only"
    TOAST = "toast"


class AdminUISettings(FidesSettings):
    """Configuration settings for the Admin UI."""

    enabled: bool = Field(
        default=True, description="Toggle whether the Admin UI is served."
    )
    url: SerializeAsAny[Optional[AnyHttpUrlStringRemovesSlash]] = Field(
        default=None, description="The base URL for the Admin UI."
    )
    max_privacy_request_download_rows: int = Field(
        default=100000,
        description="The maximum number of rows permitted to be returned in a privacy request report download",
    )
    error_notification_mode: Optional[str] = Field(
        default=ErrorNotificationMode.CONSOLE_ONLY.value,
        description="This setting controls how errors are notified to users.",
    )
    model_config = SettingsConfigDict(env_prefix="FIDES__ADMIN_UI__")



"""Add fides_user_respondent_email_verification table

Revision ID: 440a5b9a3493
Revises: 2263583b0e44
Create Date: 2025-05-29 15:41:59.521940

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "440a5b9a3493"
down_revision = "2263583b0e44"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "fides_user_respondent_email_verification",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column("user_id", sa.String(), nullable=False),
        sa.Column("access_token", sa.String(), nullable=False),
        sa.Column(
            "access_token_expires_at", sa.DateTime(timezone=True), nullable=False
        ),
        sa.Column("identity_verified_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            nullable=False,
            server_default=sa.text("now()"),
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            nullable=False,
            server_default=sa.text("now()"),
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["fidesuser.id"],
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        "ix_fides_user_respondent_email_verification_id",
        "fides_user_respondent_email_verification",
        ["id"],
        unique=True,
    )
    op.create_index(
        "ix_fides_user_respondent_email_verification_access_token",
        "fides_user_respondent_email_verification",
        ["access_token"],
        unique=True,
    )
    op.create_index(
        "ix_fides_user_respondent_email_verification_user_id",
        "fides_user_respondent_email_verification",
        ["user_id"],
    )


def downgrade():
    op.drop_index(
        "ix_fides_user_respondent_email_verification_user_id",
        table_name="fides_user_respondent_email_verification",
    )
    op.drop_index(
        "ix_fides_user_respondent_email_verification_access_token",
        table_name="fides_user_respondent_email_verification",
    )
    op.drop_index(
        "ix_fides_user_respondent_email_verification_id",
        table_name="fides_user_respondent_email_verification",
    )
    op.drop_table("fides_user_respondent_email_verification")



"""
Fides-specific extensions to the pydantic models of taxonomy elements as defined in fideslang.
"""

from typing import List, Optional

from fideslang.models import DataCategory as BaseDataCategory
from fideslang.models import DataSubject as BaseDataSubject
from fideslang.models import DataSubjectRights
from fideslang.models import DataUse as BaseDataUse
from fideslang.models import DefaultModel
from fideslang.validation import FidesKey
from pydantic import BaseModel, Field

active_field = Field(
    default=True, description="Indicates whether the resource is currently 'active'."
)


class DataUse(BaseDataUse):
    active: bool = active_field


class DataCategory(BaseDataCategory):
    active: bool = active_field


class DataSubject(BaseDataSubject):
    active: bool = active_field


class TaxonomyCreateOrUpdateBase(DefaultModel, BaseModel):
    name: Optional[str] = None
    description: Optional[str] = None
    active: Optional[bool] = True
    fides_key: Optional[FidesKey] = None
    tags: Optional[List[str]] = None
    organization_fides_key: Optional[FidesKey] = "default_organization"


class DataUseCreateOrUpdate(TaxonomyCreateOrUpdateBase):
    parent_key: Optional[FidesKey] = None


class DataCategoryCreateOrUpdate(TaxonomyCreateOrUpdateBase):
    parent_key: Optional[FidesKey] = None


class DataSubjectCreateOrUpdate(TaxonomyCreateOrUpdateBase):
    rights: Optional[DataSubjectRights] = None
    automated_decisions_or_profiling: Optional[bool] = None



import secrets
from typing import Any, Dict, List, Optional, TypeVar

from loguru import logger

from fides.api.db.session import get_db_session
from fides.api.models.masking_secret import MaskingSecret
from fides.api.schemas.masking.masking_secrets import (
    MaskingSecretCache,
    MaskingSecretMeta,
    SecretType,
)
from fides.api.util.cache import get_cache, get_masking_secret_cache_key
from fides.config import CONFIG

T = TypeVar("T")


class SecretsUtil:
    @staticmethod
    def get_or_generate_secret(
        privacy_request_id: Optional[str],
        secret_type: SecretType,
        masking_secret_meta: MaskingSecretMeta[T],
    ) -> Optional[T]:
        if privacy_request_id is not None:
            secret = SecretsUtil.get_masking_secret(
                privacy_request_id=privacy_request_id,
                masking_strategy=masking_secret_meta.masking_strategy,
                secret_type=secret_type,
            )
            if not secret:
                logger.warning(
                    "Secret type {} expected but was not present for masking strategy {}",
                    secret_type,
                    masking_secret_meta.masking_strategy,
                )
            return secret

        # expected for standalone masking service
        return masking_secret_meta.generate_secret_func(
            masking_secret_meta.secret_length
        )

    @staticmethod
    def generate_secret_string(length: int) -> str:
        return secrets.token_urlsafe(length)

    @staticmethod
    def generate_secret_bytes(length: int) -> bytes:
        return secrets.token_bytes(length)

    @staticmethod
    def build_masking_secrets_for_cache(
        masking_secret_meta: Dict[SecretType, MaskingSecretMeta[T]],
    ) -> List[MaskingSecretCache[T]]:
        masking_secrets = []
        for secret_type in masking_secret_meta.keys():
            meta = masking_secret_meta[secret_type]
            secret: T = meta.generate_secret_func(meta.secret_length)
            masking_secrets.append(
                MaskingSecretCache[T](  # type: ignore
                    secret=secret,
                    masking_strategy=meta.masking_strategy,
                    secret_type=secret_type,
                )
            )
        return masking_secrets

    @staticmethod
    def get_masking_secret(
        privacy_request_id: str,
        masking_strategy: str,
        secret_type: SecretType,
    ) -> Optional[Any]:
        """
        Attempts to retrieve masking secret from cache first, then falls back to DB.
        """
        # TODO: get rid of the cache check after a few releases
        # Try cache first
        cache = get_cache()
        cache_key = get_masking_secret_cache_key(
            privacy_request_id=privacy_request_id,
            masking_strategy=masking_strategy,
            secret_type=secret_type,
        )
        secret = cache.get_encoded_by_key(cache_key)
        if secret is not None:
            return secret

        # Cache miss - try database
        session_local = get_db_session(CONFIG)
        with session_local() as session:
            masking_secret: MaskingSecret = (
                session.query(MaskingSecret)
                .filter(
                    MaskingSecret.privacy_request_id == privacy_request_id,
                    MaskingSecret.masking_strategy == masking_strategy,
                    MaskingSecret.secret_type == secret_type,
                )
                .first()
            )

            if not masking_secret:
                return None

            return masking_secret.get_secret()



from sqlalchemy.orm import Session

from fides.api.graph.config import (
    Collection,
    CollectionAddress,
    Field,
    GraphDataset,
    ScalarField,
)
from fides.api.graph.graph import Node
from fides.api.graph.traversal import TraversalNode
from fides.api.models.connectionconfig import ConnectionConfig

# Import application models
from fides.api.models.manual_task import (
    ManualTask,
    ManualTaskConfig,
    ManualTaskConfigurationType,
    ManualTaskEntityType,
    ManualTaskInstance,
)
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.schemas.policy import ActionType
from fides.api.task.manual.manual_task_address import ManualTaskAddress


def get_connection_configs_with_manual_tasks(db: Session) -> list[ConnectionConfig]:
    """
    Get all connection configs that have manual tasks.
    """
    return (
        db.query(ConnectionConfig)
        .join(ManualTask, ConnectionConfig.id == ManualTask.parent_entity_id)
        .filter(ManualTask.parent_entity_type == "connection_config")
        .filter(ConnectionConfig.disabled.is_(False))
        .all()
    )


def get_manual_task_addresses(db: Session) -> list[CollectionAddress]:
    """
    Get manual task addresses for all connection configs that have manual tasks.

    Note: Manual tasks should be included in the graph if they exist for any connection config
    that's part of the dataset graph, regardless of specific policy targets. This allows
    manual tasks to collect additional data that may be needed for the privacy request.
    """
    # Get all connection configs that have manual tasks (excluding disabled ones)
    connection_configs_with_manual_tasks = get_connection_configs_with_manual_tasks(db)

    # Create addresses for all connections that have manual tasks
    manual_task_addresses = []
    for config in connection_configs_with_manual_tasks:
        manual_task_addresses.append(ManualTaskAddress.create(config.key))

    return manual_task_addresses


def get_manual_task_for_connection_config(
    db: Session, connection_config_key: str
) -> ManualTask:
    """Get the ManualTask for a specific connection config,
    the manual task/connection config relationship is 1:1.
    """
    return (
        db.query(ManualTask)
        .join(ConnectionConfig, ManualTask.parent_entity_id == ConnectionConfig.id)
        .filter(
            ConnectionConfig.key == connection_config_key,
            ManualTask.parent_entity_type == "connection_config",
        )
        .one_or_none()
    )


def create_manual_data_traversal_node(
    db: Session, address: CollectionAddress
) -> "TraversalNode":
    """
    Create a TraversalNode for a manual_data collection
    """
    connection_key = address.dataset

    # Get manual tasks for this connection to determine fields
    manual_task = get_manual_task_for_connection_config(db, connection_key)

    # Create fields based on ManualTaskConfigFields
    fields: list[Field] = []
    for config in manual_task.configs:
        for field in config.field_definitions:
            # Create a scalar field for each manual task field
            # Extract data categories from field metadata if available
            field_metadata = field.field_metadata or {}
            data_categories = field_metadata.get("data_categories", [])

            scalar_field = ScalarField(
                name=field.field_key,
                data_categories=data_categories,
                # Manual task fields don't have complex relationships
            )
            fields.append(scalar_field)

    # Create a synthetic Collection
    collection = Collection(
        name=ManualTaskAddress.MANUAL_DATA_COLLECTION,
        fields=fields,
        # Manual tasks don't have complex dependencies
        after=set(),
    )

    # Create a synthetic GraphDataset
    dataset = GraphDataset(
        name=connection_key,
        collections=[collection],
        connection_key=connection_key,
        after=set(),
    )

    node = Node(dataset, collection)
    traversal_node = TraversalNode(node)

    return traversal_node


def create_manual_task_instances_for_privacy_request(
    db: Session, privacy_request: PrivacyRequest
) -> list[ManualTaskInstance]:
    """Create ManualTaskInstance entries for all active manual tasks relevant to a privacy request."""
    instances = []

    # Get all connection configs that have manual tasks (excluding disabled ones)
    connection_configs_with_manual_tasks = get_connection_configs_with_manual_tasks(db)

    # Determine the privacy request type based on policy rules
    has_access_rules = bool(
        privacy_request.policy.get_rules_for_action(action_type=ActionType.access)
    )
    has_erasure_rules = bool(
        privacy_request.policy.get_rules_for_action(action_type=ActionType.erasure)
    )

    for connection_config in connection_configs_with_manual_tasks:
        manual_tasks = (
            db.query(ManualTask)
            .filter(
                ManualTask.parent_entity_id == connection_config.id,
                ManualTask.parent_entity_type == "connection_config",
            )
            .all()
        )

        for manual_task in manual_tasks:
            # Get the active config for this manual task, filtered by request type
            active_config_query = db.query(ManualTaskConfig).filter(
                ManualTaskConfig.task_id == manual_task.id,
                ManualTaskConfig.is_current.is_(True),
            )

            # Filter by configuration type based on privacy request type
            if has_access_rules and has_erasure_rules:
                # If both access and erasure rules exist, include both types
                active_config_query = active_config_query.filter(
                    ManualTaskConfig.config_type.in_(
                        [
                            ManualTaskConfigurationType.access_privacy_request,
                            ManualTaskConfigurationType.erasure_privacy_request,
                        ]
                    )
                )
            elif has_access_rules:
                # Only access rules - only include access configurations
                active_config_query = active_config_query.filter(
                    ManualTaskConfig.config_type
                    == ManualTaskConfigurationType.access_privacy_request
                )
            elif has_erasure_rules:
                # Only erasure rules - only include erasure configurations
                active_config_query = active_config_query.filter(
                    ManualTaskConfig.config_type
                    == ManualTaskConfigurationType.erasure_privacy_request
                )
            else:
                # No relevant rules - skip this manual task
                continue

            active_configs = active_config_query.all()

            if not active_configs:
                continue  # Skip if no active configs

            # Create instances for each active config
            for active_config in active_configs:
                # Check if instance already exists for this config
                existing_instance = (
                    db.query(ManualTaskInstance)
                    .filter(
                        ManualTaskInstance.entity_id == privacy_request.id,
                        ManualTaskInstance.entity_type == "privacy_request",
                        ManualTaskInstance.task_id == manual_task.id,
                        ManualTaskInstance.config_id == active_config.id,
                    )
                    .first()
                )

                if not existing_instance:
                    instance = ManualTaskInstance(
                        entity_id=privacy_request.id,
                        entity_type=ManualTaskEntityType.privacy_request,
                        task_id=manual_task.id,
                        config_id=active_config.id,
                    )
                    db.add(instance)
                    instances.append(instance)

    if instances:
        db.commit()

    return instances


def get_manual_task_instances_for_privacy_request(
    db: Session, privacy_request: PrivacyRequest
) -> list[ManualTaskInstance]:
    """Get all manual task instances for a privacy request."""
    return (
        db.query(ManualTaskInstance)
        .filter(
            ManualTaskInstance.entity_id == privacy_request.id,
            ManualTaskInstance.entity_type == "privacy_request",
        )
        .all()
    )


def create_manual_task_artificial_graphs(
    db: Session,
) -> list:
    """
    Create artificial GraphDataset objects for manual tasks that can be included
    in the main dataset graph during the dataset configuration phase.

    Manual tasks should be treated as data sources/datasets rather than being
    appended to the traversal graph later.

    Manual task collections are designed as root nodes that execute immediately when
    the privacy request starts, in parallel with identity processing. They don't depend
    on identity data since they provide manually-entered data rather than consuming it.

    Args:
        db: Database session
        policy: The policy being executed (optional, for filtering manual task configs)

    Returns:
        List of GraphDataset objects representing manual tasks as root nodes
    """

    manual_task_graphs = []
    manual_addresses = get_manual_task_addresses(db)

    for address in manual_addresses:
        connection_key = address.dataset

        # Get manual tasks for this connection to determine fields
        manual_task = get_manual_task_for_connection_config(db, connection_key)

        # Create fields based only on ManualTaskConfigFields
        fields: list = []

        # Manual task collections act as root nodes - they don't need identity dependencies
        # since they provide manually-entered data rather than consuming identity data.
        current_configs = [
            config for config in manual_task.configs if config.is_current
        ]
        for config in current_configs:
            if config.config_type not in [
                ManualTaskConfigurationType.access_privacy_request,
                ManualTaskConfigurationType.erasure_privacy_request,
            ]:
                continue

            for field in config.field_definitions:
                # Create a scalar field for each manual task field
                field_metadata = field.field_metadata or {}
                data_categories = field_metadata.get("data_categories", [])

                scalar_field = ScalarField(
                    name=field.field_key,
                    data_categories=data_categories,
                )
                fields.append(scalar_field)

        if fields:  # Only create graph if there are fields
            # Create a synthetic Collection
            collection = Collection(
                name=ManualTaskAddress.MANUAL_DATA_COLLECTION,
                fields=fields,
                # Manual tasks have no dependencies - they're root nodes
                after=set(),
            )

            # Create a synthetic GraphDataset
            graph_dataset = GraphDataset(
                name=connection_key,
                collections=[collection],
                connection_key=connection_key,
                after=set(),
            )

            manual_task_graphs.append(graph_dataset)

    return manual_task_graphs



"""add enabled_actions to connectionconfig

Revision ID: e798f37f0c26
Revises: 317e6197c76a
Create Date: 2023-05-11 17:54:05.476225

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "e798f37f0c26"
down_revision = "317e6197c76a"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "connectionconfig",
        sa.Column(
            "enabled_actions",
            postgresql.ARRAY(sa.String()),
            nullable=True,
        ),
    )


def downgrade():
    op.drop_column("connectionconfig", "enabled_actions")



# pylint: disable=no-name-in-module
"""Logic related to sanitizing and validating user application input."""
from html import escape
from re import compile as regex
from typing import Annotated

from nh3 import clean
from pydantic import AfterValidator, AnyHttpUrl, AnyUrl, BeforeValidator

from fides.api.util.unsafe_file_util import verify_css


def validate_safe_str(val: str) -> str:
    """
    Validator for SafeStr that is designed to be used in place of the `str` type
    any place where user input is expected.

    The validation applied here does the typical sanitization/cleanup
    that is required when dealing with user-supplied input.
    """
    # HTML Escapes
    value = escape(val)

    if len(value) > 32000:
        raise ValueError("Value must be 32000 characters or less.")

    return value


SafeStr = Annotated[str, BeforeValidator(validate_safe_str)]


def validate_html_str(val: str) -> str:
    """
    Validator function for HTMLStr - designed to be used in place of the `str` type
    any place where user inputted HTML text is expected.

    The validation applied here enforces that only a subset of "safe" HTML is
    supported to prevent XSS or similar attacks.

    """
    # Assert text doesn't include an complex/malicious HTML.
    # Allow only basic markup tags, for extra safety
    ALLOWED_HTML_TAGS = {
        "a",
        "blockquote",
        "br",
        "div",
        "em",
        "h1",
        "h2",
        "h3",
        "h4",
        "h5",
        "h6",
        "i",
        "li",
        "ol",
        "p",
        "pre",
        "q",
        "rt",
        "ruby",
        "s",
        "span",
        "strong",
        "u",
    }

    ALLOWED_ATTRIBUTES = {
        "a": {"href", "hreflang", "target"},
    }

    if val:
        return clean(val, tags=ALLOWED_HTML_TAGS, attributes=ALLOWED_ATTRIBUTES)
    return val


HtmlStr = Annotated[str, BeforeValidator(validate_html_str)]


def validate_phone_number(value: str) -> str:
    """Validator for PhoneNumber

    Standard format can be found here: https://en.wikipedia.org/wiki/E.164
    """
    # The front-end sends an empty string if the user doesn't input anything
    if value == "":
        return ""
    max_length = 16  # Includes the +
    min_length = 9
    pattern = regex(r"^\+[1-9]\d{1,14}$")
    if len(value) > max_length or len(value) < min_length or not pattern.search(value):
        raise ValueError(
            "Phone number must be formatted in E.164 format, i.e. '+15558675309'."
        )
    return value


PhoneNumber = Annotated[str, BeforeValidator(validate_phone_number)]


def validate_gpp_mechanism_consent_value(value: str) -> str:
    """
    Validator for GPPMechanismConsentValue -  Allowable consent values for GPP Mechanism Mappings.
    """
    pattern = regex(r"^\d+$")
    if not isinstance(value, str) or not pattern.search(value):
        raise ValueError("GPP Mechanism consent value must be a string of digits.")
    return value


GPPMechanismConsentValue = Annotated[
    str, BeforeValidator(validate_gpp_mechanism_consent_value)
]


def validate_path_of_url(value: AnyUrl) -> str:
    """
    A URL origin value. See https://developer.mozilla.org/en-US/docs/Glossary/Origin.

    We perform the basic URL validation, but also prevent URLs with paths,
    as paths are not part of an origin.

    As this is meant to be used as a validator _after_ the AnyUrl validator,
    we implicitly disallow `*` to be specified as an origin value, since it is
    rejected by the AnyUrl validator.

    Note that `*` _is_ considered a valid origin value
    (https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Origin),
    but we specifically disallow it to be set on the `cors_origins` security setting
    in our application. This is because non-owner users (contributors)
    are able to edit the `cors_origins` security setting (via API), and we do not
    want to allow them to set a wildcard origin.

    Instead, `.*` can be set via the `cors_origin_regex` security setting
    to effectively allow all origins. This is roughly equivalent to setting
    `"*"` as a `cors_origin`, but there is a slight behavioral difference:
    the `Access-Control-Allow-Origin` in the response will be set to the
    `Access-Control-Allow-Origin` specified in the request rather than `"*"`.
    """
    if value.path and value.path != "/":
        raise ValueError("URL origin values cannot contain a path.")

    # Intentionally serializing as a string instead of a URL.
    # Intentionally removing trailing slash. In Pydantic V2 AnyURL will add a trailing slash to root URL's so stripping this off.
    return str(value).rstrip("/")


# This custom type is used to validate HTTP origins, e.g. CORS origins
# used in SecuritySettings.  See docstring for `validate_path_of_url`
# for more details.
URLOriginString = Annotated[
    AnyUrl,
    AfterValidator(validate_path_of_url),
]


def validate_css_str(value: str) -> str:
    """
    Validator for CssStr

    A custom string type that represents a valid CSS stylesheet.

    The `CssStr` type automatically validates the input value using the `verify_css` function
    to ensure that it is a valid CSS stylesheet. The validation checks for parsing errors,
    import statements, and calls to the url() function.

    If the input value is not a string or fails the CSS validation, a `ValidationError` is raised.
    """
    if not isinstance(value, str):
        raise TypeError("CssStr must be a string")
    verify_css(value)
    return value


CssStr = Annotated[str, BeforeValidator(validate_css_str)]


def validate_path_of_http_url_no_slash(value: AnyHttpUrl) -> str:
    """Converts an AnyHttpUrl to a string and strips trailing slash"""
    return str(value).rstrip("/")


AnyHttpUrlStringRemovesSlash = Annotated[
    AnyHttpUrl, AfterValidator(validate_path_of_http_url_no_slash)
]



import secrets
from typing import Optional, Set, Union

from loguru import logger
from sqlalchemy.orm import Session

from fides.api.common_exceptions import (
    IdentityNotFoundException,
    PolicyNotFoundException,
)
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import (
    ConsentRequest,
    PrivacyRequest,
    PrivacyRequestError,
    PrivacyRequestNotifications,
    ProvidedIdentityType,
)
from fides.api.schemas.messaging.messaging import (
    EMAIL_MESSAGING_SERVICES,
    ErrorNotificationBodyParams,
    FidesopsMessage,
    MessagingActionType,
    RequestReceiptBodyParams,
    RequestReviewDenyBodyParams,
    SubjectIdentityVerificationBodyParams,
)
from fides.api.schemas.policy import ActionType
from fides.api.schemas.redis_cache import Identity
from fides.api.service.messaging.message_dispatch_service import (
    EMAIL_JOIN_STRING,
    dispatch_message,
    dispatch_message_task,
    message_send_enabled,
)
from fides.api.tasks import MESSAGING_QUEUE_NAME
from fides.config import FidesConfig, get_config
from fides.config.config_proxy import ConfigProxy


class MessagingService:
    def __init__(self, db: Session, config: FidesConfig, config_proxy: ConfigProxy):
        self.db = db
        self.config = config
        self.config_proxy = config_proxy

    def send_request_approved(self, privacy_request: PrivacyRequest) -> None:
        if message_send_enabled(
            self.db,
            privacy_request.property_id,
            MessagingActionType.PRIVACY_REQUEST_REVIEW_APPROVE,
            self.config_proxy.notifications.send_request_review_notification,
        ):
            identity_data = privacy_request.get_cached_identity_data()
            if not identity_data:
                logger.error(
                    IdentityNotFoundException(
                        "Identity was not found, so request review message could not be sent."
                    )
                )
                return

            to_identity = Identity(
                email=identity_data.get(ProvidedIdentityType.email.value),
                phone_number=identity_data.get(ProvidedIdentityType.phone_number.value),
            )

            dispatch_message_task.apply_async(
                queue=MESSAGING_QUEUE_NAME,
                kwargs={
                    "message_meta": FidesopsMessage(
                        action_type=MessagingActionType.PRIVACY_REQUEST_REVIEW_APPROVE,
                        body_params=None,
                    ).model_dump(mode="json"),
                    "service_type": self.config_proxy.notifications.notification_service_type,
                    "to_identity": to_identity.model_dump(mode="json"),
                    "property_id": privacy_request.property_id,
                },
            )

    def send_request_denied(
        self, privacy_request: PrivacyRequest, deny_reason: Optional[str] = None
    ) -> None:
        if message_send_enabled(
            self.db,
            privacy_request.property_id,
            MessagingActionType.PRIVACY_REQUEST_REVIEW_DENY,
            self.config_proxy.notifications.send_request_review_notification,
        ):
            identity_data = privacy_request.get_cached_identity_data()
            if not identity_data:
                logger.error(
                    IdentityNotFoundException(
                        "Identity was not found, so request review message could not be sent."
                    )
                )
                return

            to_identity = Identity(
                email=identity_data.get(ProvidedIdentityType.email.value),
                phone_number=identity_data.get(ProvidedIdentityType.phone_number.value),
            )

            dispatch_message_task.apply_async(
                queue=MESSAGING_QUEUE_NAME,
                kwargs={
                    "message_meta": FidesopsMessage(
                        action_type=MessagingActionType.PRIVACY_REQUEST_REVIEW_DENY,
                        body_params=RequestReviewDenyBodyParams(
                            rejection_reason=deny_reason
                        ),
                    ).model_dump(mode="json"),
                    "service_type": self.config_proxy.notifications.notification_service_type,
                    "to_identity": to_identity.model_dump(mode="json"),
                    "property_id": privacy_request.property_id,
                },
            )

    def send_verification_code(
        self,
        request: Union[ConsentRequest, PrivacyRequest],
        to_identity: Optional[Identity],
        property_id: Optional[str],
    ) -> Optional[str]:

        verification_code = None

        if message_send_enabled(
            self.db,
            request.property_id,
            MessagingActionType.SUBJECT_IDENTITY_VERIFICATION,
            self.config_proxy.execution.subject_identity_verification_required,
        ):
            verification_code = _generate_id_verification_code()
            request.cache_identity_verification_code(verification_code)

            dispatch_message(
                self.db,
                action_type=MessagingActionType.SUBJECT_IDENTITY_VERIFICATION,
                to_identity=to_identity,
                service_type=self.config_proxy.notifications.notification_service_type,
                message_body_params=SubjectIdentityVerificationBodyParams(
                    verification_code=verification_code,
                    verification_code_ttl_seconds=self.config.redis.identity_verification_code_ttl_seconds,
                ),
                property_id=property_id,
            )

        return verification_code

    def send_privacy_request_receipt(
        self,
        policy: Optional[Policy],
        identity: Identity,
        privacy_request: PrivacyRequest,
    ) -> None:
        if message_send_enabled(
            self.db,
            privacy_request.property_id,
            MessagingActionType.PRIVACY_REQUEST_RECEIPT,
            self.config_proxy.notifications.send_request_receipt_notification,
        ):
            if not identity:
                logger.error(
                    IdentityNotFoundException(
                        "Identity was not found, so request receipt message could not be sent."
                    )
                )
                return

            if not policy:
                logger.error(
                    PolicyNotFoundException(
                        "Policy was not found, so request receipt message could not be sent."
                    )
                )
                return

            request_types: Set[str] = set()
            for action_type in ActionType:
                if policy.get_rules_for_action(action_type=ActionType(action_type)):
                    request_types.add(action_type)

            request_types_list = list(request_types)

            dispatch_message_task.apply_async(
                queue=MESSAGING_QUEUE_NAME,
                kwargs={
                    "message_meta": FidesopsMessage(
                        action_type=MessagingActionType.PRIVACY_REQUEST_RECEIPT,
                        body_params=RequestReceiptBodyParams(
                            request_types=request_types_list
                        ),
                    ).model_dump(mode="json"),
                    "service_type": self.config.notifications.notification_service_type,
                    "to_identity": identity.labeled_dict(),
                    "property_id": privacy_request.property_id,
                },
            )


def _generate_id_verification_code() -> str:
    """
    Generate one-time identity verification code
    """
    return str(secrets.choice(range(100000, 999999)))


def check_and_dispatch_error_notifications(db: Session) -> None:
    config_proxy = ConfigProxy(db)
    privacy_request_notifications = PrivacyRequestNotifications.all(db=db)
    if not privacy_request_notifications:
        return None

    unsent_errors = PrivacyRequestError.filter(
        db=db, conditions=(PrivacyRequestError.message_sent.is_(False))
    ).all()
    if not unsent_errors:
        return None

    email_config = (
        config_proxy.notifications.notification_service_type in EMAIL_MESSAGING_SERVICES
    )

    if (
        email_config
        and len(unsent_errors) >= privacy_request_notifications[0].notify_after_failures
    ):
        for email in privacy_request_notifications[0].email.split(EMAIL_JOIN_STRING):
            dispatch_message_task.apply_async(
                queue=MESSAGING_QUEUE_NAME,
                kwargs={
                    "message_meta": FidesopsMessage(
                        action_type=MessagingActionType.PRIVACY_REQUEST_ERROR_NOTIFICATION,
                        body_params=ErrorNotificationBodyParams(
                            unsent_errors=len(unsent_errors)
                        ),
                    ).model_dump(mode="json"),
                    "service_type": config_proxy.notifications.notification_service_type,
                    "to_identity": {"email": email},
                    "property_id": None,
                },
            )

        for error in unsent_errors:
            error.update(db=db, data={"message_sent": True})

    return None


def send_verification_code_to_user(
    db: Session,
    request: Union[ConsentRequest, PrivacyRequest],
    to_identity: Optional[Identity],
    property_id: Optional[str],
) -> str:
    """Generate and cache a verification code, and then message the user"""
    config = get_config()
    config_proxy = ConfigProxy(db)
    verification_code = _generate_id_verification_code()
    request.cache_identity_verification_code(verification_code)
    messaging_action_type = MessagingActionType.SUBJECT_IDENTITY_VERIFICATION
    dispatch_message(
        db,
        action_type=messaging_action_type,
        to_identity=to_identity,
        service_type=config_proxy.notifications.notification_service_type,
        message_body_params=SubjectIdentityVerificationBodyParams(
            verification_code=verification_code,
            verification_code_ttl_seconds=config.redis.identity_verification_code_ttl_seconds,
        ),
        property_id=property_id,
    )

    return verification_code


def send_privacy_request_receipt_message_to_user(
    policy: Optional[Policy],
    to_identity: Optional[Identity],
    service_type: Optional[str],
    property_id: Optional[str],
) -> None:
    """Helper function to send request receipt message to the user"""
    if not to_identity:
        logger.error(
            IdentityNotFoundException(
                "Identity was not found, so request receipt message could not be sent."
            )
        )
        return
    if not policy:
        logger.error(
            PolicyNotFoundException(
                "Policy was not found, so request receipt message could not be sent."
            )
        )
        return
    request_types: Set[str] = set()
    for action_type in ActionType:
        if policy.get_rules_for_action(action_type=ActionType(action_type)):
            request_types.add(action_type)

    request_types_list = list(request_types)

    dispatch_message_task.apply_async(
        queue=MESSAGING_QUEUE_NAME,
        kwargs={
            "message_meta": FidesopsMessage(
                action_type=MessagingActionType.PRIVACY_REQUEST_RECEIPT,
                body_params=RequestReceiptBodyParams(request_types=request_types_list),
            ).model_dump(mode="json"),
            "service_type": service_type,
            "to_identity": to_identity.labeled_dict(),
            "property_id": property_id,
        },
    )



"""add_execution_timeframe

Revision ID: bde646a6f51e
Revises: c2f7a29c4780
Create Date: 2022-09-01 15:46:47.954354

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "bde646a6f51e"
down_revision = "c2f7a29c4780"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "policy", sa.Column("execution_timeframe", sa.Integer(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("policy", "execution_timeframe")
    # ### end Alembic commands ###



from typing import ClassVar, List, Optional

from pydantic import Field

from fides.api.schemas.base_class import NoValidationSchema
from fides.api.schemas.connection_configuration.connection_secrets import (
    ConnectionConfigSecretsSchema,
)


class ScyllaSchema(ConnectionConfigSecretsSchema):
    """Schema to validate the secrets needed to connect to a Scylla Database"""

    host: str = Field(
        title="Host",
        description="The hostname or IP address of the server where the database is running.",
    )
    port: int = Field(
        9042,
        title="Port",
        description="The network port number on which the server is listening for incoming connections (default: 9042).",
    )
    username: str = Field(
        title="Username",
        description="The user account used to authenticate and access the database.",
    )
    password: str = Field(
        title="Password",
        description="The password used to authenticate and access the database.",
        json_schema_extra={"sensitive": True},
    )
    keyspace: Optional[str] = Field(
        default=None,
        title="Keyspace",
        description="The keyspace used. If not provided, DSRs for this integration will error. If the integration is used for D & D, then setting a keyspace is not required.",
    )

    _required_components: ClassVar[List[str]] = [
        "host",
        "username",
        "password",
    ]


class ScyllaDocsSchema(ScyllaSchema, NoValidationSchema):
    """Scylla Secrets Schema for API docs"""



"""adding masking secrets

Revision ID: 7e9a2b52f498
Revises: d0031087eacb
Create Date: 2024-12-23 23:29:53.557951

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op

# revision identifiers, used by Alembic.
revision = "7e9a2b52f498"
down_revision = "d0031087eacb"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "masking_secret",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("privacy_request_id", sa.String(), nullable=False),
        sa.Column(
            "secret",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=False,
        ),
        sa.Column("masking_strategy", sa.String(), nullable=False),
        sa.Column(
            "secret_type",
            sa.String(),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["privacy_request_id"], ["privacyrequest.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_masking_secret_id"), "masking_secret", ["id"], unique=False
    )


def downgrade():
    op.drop_index(op.f("ix_masking_secret_id"), table_name="masking_secret")
    op.drop_table("masking_secret")



# pylint: disable=too-many-lines
import json
from typing import Any, Dict, List, Optional, Set

import networkx
from loguru import logger
from networkx import NetworkXNoCycle
from sqlalchemy.orm import Query, Session

from fides.api.common_exceptions import TraversalError
from fides.api.graph.config import (
    ROOT_COLLECTION_ADDRESS,
    TERMINATOR_ADDRESS,
    CollectionAddress,
    FieldAddress,
)
from fides.api.graph.graph import DatasetGraph
from fides.api.graph.traversal import (
    ARTIFICIAL_NODES,
    Traversal,
    TraversalNode,
    log_traversal_error_and_update_privacy_request,
)
from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import (
    COMPLETED_EXECUTION_LOG_STATUSES,
    PrivacyRequest,
    RequestTask,
    TraversalDetails,
)
from fides.api.models.worker_task import ExecutionLogStatus
from fides.api.schemas.policy import ActionType
from fides.api.task.deprecated_graph_task import format_data_use_map_for_caching
from fides.api.task.execute_request_tasks import log_task_queued, queue_request_task
from fides.api.task.manual.manual_task_address import ManualTaskAddress
from fides.api.task.manual.manual_task_utils import (
    create_manual_task_instances_for_privacy_request,
)
from fides.api.util.logger_context_utils import log_context


def _add_edge_if_no_nodes(
    traversal_nodes: Dict[CollectionAddress, TraversalNode],
    networkx_graph: networkx.DiGraph,
) -> None:
    """
    Adds an edge from the root node to the terminator node, altering the networkx_graph in-place.

    Handles edge case if there are no traversal nodes in the graph at all
    """
    if not traversal_nodes.items():
        networkx_graph.add_edge(ROOT_COLLECTION_ADDRESS, TERMINATOR_ADDRESS)


def build_access_networkx_digraph(
    traversal_nodes: Dict[CollectionAddress, TraversalNode],
    end_nodes: List[CollectionAddress],
    traversal: Traversal,
) -> networkx.DiGraph:
    """
    DSR 3.0: Builds an access networkx graph to get consistent formatting of nodes to build the Request Tasks,
    regardless of whether node is real or artificial.

    Primarily though, this lets us use networkx.descendants to calculate every node that can be reached from the current
    node to more easily mark downstream nodes as failed if the current node fails.
    """
    networkx_graph = networkx.DiGraph()
    networkx_graph.add_nodes_from(traversal_nodes.keys())
    networkx_graph.add_nodes_from(ARTIFICIAL_NODES)

    # The first nodes visited are the nodes that only need identity data.
    # Therefore, they are all immediately downstream of the root.
    first_nodes: Dict[FieldAddress, str] = traversal.extract_seed_field_addresses()

    for node in [
        CollectionAddress(initial_node.dataset, initial_node.collection)
        for initial_node in first_nodes
    ]:
        networkx_graph.add_edge(ROOT_COLLECTION_ADDRESS, node)

    for collection_address, traversal_node in traversal_nodes.items():
        for child in traversal_node.children:
            # For every node, add a downstream edge to its children
            # that were calculated in traversal.traverse
            networkx_graph.add_edge(collection_address, child)

    for node in end_nodes:
        # Connect the end nodes, those that have no downstream dependencies, to the terminator node
        networkx_graph.add_edge(node, TERMINATOR_ADDRESS)

    manual_nodes = [
        addr
        for addr in traversal_nodes.keys()
        if addr.collection == ManualTaskAddress.MANUAL_DATA_COLLECTION
    ]
    for manual_node in manual_nodes:
        networkx_graph.add_edge(ROOT_COLLECTION_ADDRESS, manual_node)

    _add_edge_if_no_nodes(traversal_nodes, networkx_graph)
    return networkx_graph


def _evaluate_erasure_dependencies(
    traversal_node: TraversalNode, end_nodes: List[CollectionAddress]
) -> Set[CollectionAddress]:
    """
    Return a set of collection addresses corresponding to collections that need
    to be erased before the given task.

    Remove the dependent collection addresses
    from `end_nodes` so they can be executed in the correct order. If a task does
    not have any dependencies it is linked directly to the root node
    """
    erase_after = traversal_node.node.collection.erase_after
    for collection in erase_after:
        if collection in end_nodes:
            # end_node list is modified in place
            end_nodes.remove(collection)
    # this task will execute after the collections in `erase_after` or
    # execute at the beginning by linking it to the root node
    if len(erase_after):
        erase_after.add(ROOT_COLLECTION_ADDRESS)
    return erase_after if len(erase_after) else {ROOT_COLLECTION_ADDRESS}


def build_erasure_networkx_digraph(
    traversal_nodes: Dict[CollectionAddress, TraversalNode],
    end_nodes: List[CollectionAddress],
) -> networkx.DiGraph:
    """
    DSR 3.0: Builds a networkx graph of erasure nodes to get consistent formatting of nodes to build the Request Tasks,
    regardless of whether node is real or artificial.

    Erasure graphs are different from access graphs, in that we've queried all the data we need upfront in the access
    graphs, so that all nodes can in theory run entirely in parallel, except for the "erase_after" dependencies.

    We tack on the "erase_after" dependencies here that aren't captured in traversal.traverse.

    """
    networkx_graph = networkx.DiGraph()
    networkx_graph.add_nodes_from(traversal_nodes.keys())
    networkx_graph.add_nodes_from(ARTIFICIAL_NODES)

    for node_name, traversal_node in traversal_nodes.items():
        # Add an edge from the root node to the current node, unless explicit erasure
        # dependencies are defined. Modifies end_nodes in place
        erasure_dependencies: Set[CollectionAddress] = _evaluate_erasure_dependencies(
            traversal_node, end_nodes
        )
        for dep in erasure_dependencies:
            networkx_graph.add_edge(dep, node_name)

    for node in end_nodes:
        # Connect each end node without downstream dependencies to the terminator node
        networkx_graph.add_edge(node, TERMINATOR_ADDRESS)

    try:
        # Run extra checks on the graph since we potentially modified traversal_nodes
        networkx.find_cycle(networkx_graph, ROOT_COLLECTION_ADDRESS)
    except NetworkXNoCycle:
        logger.info("No cycles found as expected")
    else:
        raise TraversalError(
            "The values for the `erase_after` fields created a cycle in the DAG."
        )

    _add_edge_if_no_nodes(traversal_nodes, networkx_graph)
    return networkx_graph


def build_consent_networkx_digraph(
    traversal_nodes: Dict[CollectionAddress, TraversalNode],
) -> networkx.DiGraph:
    """
    DSR 3.0: Builds a networkx graph of consent nodes to get consistent formatting of nodes to build the Request Tasks,
    regardless of whether node is real or artificial.
    """
    networkx_graph = networkx.DiGraph()
    networkx_graph.add_nodes_from(traversal_nodes.keys())
    networkx_graph.add_nodes_from([TERMINATOR_ADDRESS, ROOT_COLLECTION_ADDRESS])

    for collection_address, _ in traversal_nodes.items():
        # Consent graphs are simple. One node for every dataset (which has a mocked collection)
        # and no dependencies between nodes.
        networkx_graph.add_edge(ROOT_COLLECTION_ADDRESS, collection_address)
        networkx_graph.add_edge(collection_address, TERMINATOR_ADDRESS)

    _add_edge_if_no_nodes(traversal_nodes, networkx_graph)
    return networkx_graph


def base_task_data(
    graph: networkx.DiGraph,
    dataset_graph: DatasetGraph,
    privacy_request: PrivacyRequest,
    node: CollectionAddress,
    traversal_nodes: Dict[CollectionAddress, TraversalNode],
) -> Dict:
    """Build a dictionary of common RequestTask attributes that are shared for building
    access, consent, and erasure tasks"""

    collection_representation: Optional[Dict] = None
    traversal_details = {}

    if node not in ARTIFICIAL_NODES:
        # Save a representation of the collection that can be re-hydrated later
        # when executing the node, so we don't have to recalculate incoming
        # and outgoing edges.
        collection_representation = json.loads(
            # Serialize with duck typing so we get the nested sub fields as well
            dataset_graph.nodes[node].collection.model_dump_json(serialize_as_any=True)
        )

        # Saves traversal details based on data dependencies like incoming edges
        # and input keys, also useful for building the Execution Node
        if node in traversal_nodes:
            traversal_details = traversal_nodes[
                node
            ].format_traversal_details_for_save()
        else:
            # If node is not in traversal_nodes, then it is a node added for
            # custom request field processing. We manually build the traversal details,
            # with no incoming or outgoing edges and no input keys.
            graph_node = dataset_graph.nodes[node]
            traversal_details = TraversalDetails.create_empty_traversal(
                graph_node.dataset.connection_key
            ).model_dump(mode="json")

    return {
        "privacy_request_id": privacy_request.id,
        "upstream_tasks": sorted(
            [upstream.value for upstream in graph.predecessors(node)]
        ),
        "downstream_tasks": sorted(
            [downstream.value for downstream in graph.successors(node)]
        ),
        "all_descendant_tasks": sorted(
            [descend.value for descend in list(networkx.descendants(graph, node))]
        ),
        "collection_address": node.value,
        "dataset_name": node.dataset,
        "collection_name": node.collection,
        "status": (
            ExecutionLogStatus.complete
            if node == ROOT_COLLECTION_ADDRESS
            else ExecutionLogStatus.pending
        ),
        "collection": collection_representation,
        "traversal_details": traversal_details,
    }


def persist_new_access_request_tasks(
    session: Session,
    privacy_request: PrivacyRequest,
    traversal: Traversal,
    traversal_nodes: Dict[CollectionAddress, TraversalNode],
    end_nodes: List[CollectionAddress],
    dataset_graph: DatasetGraph,
) -> List[RequestTask]:
    """
    Create individual access RequestTasks from the TraversalNodes and persist to the database.
    This should only run the first time a privacy request runs.
    """
    logger.info(
        "Creating access request tasks for privacy request {}.", privacy_request.id
    )
    graph: networkx.DiGraph = build_access_networkx_digraph(
        traversal_nodes, end_nodes, traversal
    )

    for node in list(networkx.topological_sort(graph)):
        if privacy_request.get_existing_request_task(
            session, action_type=ActionType.access, collection_address=node
        ):
            continue

        RequestTask.create(
            session,
            data={
                **base_task_data(
                    graph, dataset_graph, privacy_request, node, traversal_nodes
                ),
                "access_data": (
                    [traversal.seed_data] if node == ROOT_COLLECTION_ADDRESS else []
                ),  # For consistent treatment of nodes, add the seed data to the root node.  Subsequent
                # tasks will save the data collected on the same field.
                "action_type": ActionType.access,
            },
        )

    root_task: RequestTask = privacy_request.get_root_task_by_action(ActionType.access)

    return [root_task]


def persist_initial_erasure_request_tasks(
    session: Session,
    privacy_request: PrivacyRequest,
    traversal_nodes: Dict[CollectionAddress, TraversalNode],
    end_nodes: List[CollectionAddress],
    dataset_graph: DatasetGraph,
) -> List[RequestTask]:
    """
    Create starter individual erasure RequestTasks from the TraversalNodes and persist to the database.

    These are not ready to run yet as they are still waiting for access data from the access graph
    to be able to build masking requests
    """
    logger.info(
        "Creating initial erasure request tasks for privacy request {}.",
        privacy_request.id,
    )
    graph: networkx.DiGraph = build_erasure_networkx_digraph(traversal_nodes, end_nodes)

    for node in list(networkx.topological_sort(graph)):
        if privacy_request.get_existing_request_task(
            session, action_type=ActionType.erasure, collection_address=node
        ):
            continue

        RequestTask.create(
            session,
            data={
                **base_task_data(
                    graph, dataset_graph, privacy_request, node, traversal_nodes
                ),
                "action_type": ActionType.erasure,
            },
        )

    # If a policy has an erasure rule, this method is run immediately after creating the access tasks, so their
    # nodes in the database are the same.  There are no "ready" tasks yet, because we need to wait for the
    # access step to run, so we return an empty list here.
    return []


def _get_data_for_erasures(
    session: Session, privacy_request: PrivacyRequest, request_task: RequestTask
) -> List[Dict]:
    """
    Return the access data in erasure format needed to format the masking request for the current node.
    """
    # Get the access task of the same name as the erasure task so we can transfer the data
    # collected for masking onto the current erasure task
    corresponding_access_task: Optional[RequestTask] = (
        privacy_request.get_existing_request_task(
            db=session,
            action_type=ActionType.access,
            collection_address=request_task.request_task_address,
        )
    )
    retrieved_task_data: List[Dict] = []
    if (
        corresponding_access_task
        and request_task.request_task_address not in ARTIFICIAL_NODES
    ):
        # IMPORTANT. Use "data_for_erasures" - not RequestTask.access_data.
        # For arrays, "access_data" may remove non-matched elements from arrays, but to build erasure
        # queries we need the original data in the appropriate indices
        retrieved_task_data = corresponding_access_task.get_data_for_erasures()

    return retrieved_task_data


def update_erasure_tasks_with_access_data(
    session: Session,
    privacy_request: PrivacyRequest,
) -> None:
    """
    Update individual erasure RequestTasks with data from the TraversalNodes and persist to the database.
    """
    logger.info(
        "Updating erasure request tasks with data needed for masking requests {}.",
        privacy_request.id,
    )

    for request_task in privacy_request.erasure_tasks:
        # I pull access data saved in the format suitable for erasures
        # off of the access nodes to be saved onto the erasure nodes.
        retrieved_task_data = _get_data_for_erasures(
            session, privacy_request, request_task
        )
        request_task.data_for_erasures = retrieved_task_data
        request_task.save(session)


def persist_new_consent_request_tasks(
    session: Session,
    privacy_request: PrivacyRequest,
    traversal_nodes: Dict[CollectionAddress, TraversalNode],
    identity: Dict[str, Any],
    dataset_graph: DatasetGraph,
) -> List[RequestTask]:
    """
    Create individual erasure RequestTasks from the TraversalNodes and persist to the database.  This should only
    run the first time a privacy request runs.

    Consent propagation graphs are much simpler with no relationships between nodes. Every node has identity data input,
    and every node outputs whether the consent request succeeded.
    """
    graph: networkx.DiGraph = build_consent_networkx_digraph(traversal_nodes)

    for node in list(networkx.topological_sort(graph)):
        if privacy_request.get_existing_request_task(
            session, action_type=ActionType.consent, collection_address=node
        ):
            continue
        RequestTask.create(
            session,
            data={
                **base_task_data(
                    graph, dataset_graph, privacy_request, node, traversal_nodes
                ),
                # Consent nodes take in identity data from their upstream root node
                "access_data": ([identity] if node == ROOT_COLLECTION_ADDRESS else []),
                "action_type": ActionType.consent,
            },
        )

    root_task: RequestTask = privacy_request.get_root_task_by_action(ActionType.consent)

    return [root_task]


def collect_tasks_fn(
    tn: TraversalNode, data: Dict[CollectionAddress, TraversalNode]
) -> None:
    """
    A function that is passed to traversal.traverse() that returns the modified
    traversal node with its parents and children linked as an action.
    """
    if not tn.is_root_node():
        data[tn.address] = tn


@log_context
def run_access_request(
    privacy_request: PrivacyRequest,
    policy: Policy,
    graph: DatasetGraph,
    connection_configs: List[ConnectionConfig],
    identity: Dict[str, Any],
    session: Session,
    privacy_request_proceed: bool = True,
) -> List[RequestTask]:
    """
    DSR 3.0: Build the "access" graph, add its tasks to the database and queue the root task.  If erasure rules
    are present, build the "erasure" graph at the same time so their nodes match, but these erasure nodes are
    not yet ready to run until the access graph is complete in-full.

    If we are *reprocessing* a Privacy Request, instead queue tasks whose upstream nodes are complete.
    """

    if privacy_request.access_tasks.count():
        # If we are reprocessing a privacy request, just see if there
        # are existing ready tasks; don't create new ones.
        # Possible edge cases here where we have no ready tasks and
        # Privacy Request is hanging in an in-processing state.
        ready_tasks: List[RequestTask] = get_existing_ready_tasks(
            session, privacy_request, ActionType.access
        )
    else:
        try:
            traversal: Traversal = Traversal(graph, identity, policy=policy)

            # Traversal.traverse populates traversal_nodes in place, adding parents and children to each traversal_node.
            traversal_nodes: Dict[CollectionAddress, TraversalNode] = {}
            end_nodes: List[CollectionAddress] = traversal.traverse(
                traversal_nodes, collect_tasks_fn
            )

            # Snapshot manual task field instances for this privacy request
            create_manual_task_instances_for_privacy_request(session, privacy_request)

            # Save Access Request Tasks to the database
            ready_tasks = persist_new_access_request_tasks(
                session, privacy_request, traversal, traversal_nodes, end_nodes, graph
            )

            if (
                policy.get_rules_for_action(action_type=ActionType.erasure)
                and not privacy_request.erasure_tasks.count()
            ):
                # If applicable, go ahead and save Erasure Request Tasks to the Database.
                # These erasure tasks aren't ready to run until the access graph is completed
                # in full, but this makes sure the nodes in the graphs match.
                erasure_end_nodes: List[CollectionAddress] = list(graph.nodes.keys())
                persist_initial_erasure_request_tasks(
                    session, privacy_request, traversal_nodes, erasure_end_nodes, graph
                )

            # cache a map of collections -> data uses for the output package of access requests
            privacy_request.cache_data_use_map(
                format_data_use_map_for_caching(
                    {
                        coll_address: tn.node.dataset.connection_key
                        for (coll_address, tn) in traversal_nodes.items()
                    },
                    connection_configs,
                )
            )

            # Add execution logs for skipped nodes
            if traversal.skipped_nodes:
                logger.warning(
                    "Some nodes were skipped, the identities provided were not sufficient to reach them"
                )
                for node_address, skip_message in traversal.skipped_nodes.items():
                    logger.debug(skip_message)
                    privacy_request.add_skipped_execution_log(
                        session,
                        connection_key=None,
                        dataset_name="Dataset traversal",
                        collection_name=node_address.replace(":", "."),
                        message=skip_message,
                        action_type=ActionType.access,
                    )
            # Or log success if all collections are reachable
            else:
                privacy_request.add_success_execution_log(
                    session,
                    connection_key=None,
                    dataset_name="Dataset traversal",
                    collection_name=None,
                    message=f"Traversal successful for privacy request: {privacy_request.id}",
                    action_type=ActionType.access,
                )
        except TraversalError as err:
            log_traversal_error_and_update_privacy_request(
                privacy_request, session, err
            )
            raise err

    for task in ready_tasks:
        log_task_queued(task, "main runner")
        queue_request_task(task, privacy_request_proceed)

    return ready_tasks


@log_context
def run_erasure_request(  # pylint: disable = too-many-arguments
    privacy_request: PrivacyRequest,
    session: Session,
    privacy_request_proceed: bool = True,
) -> List[RequestTask]:
    """
    DSR 3.0: Update erasure Request Tasks that were built in the "run_access_request" step with data
    collected to build masking requests and queue the root task for processing.

    If we are reprocessing a Privacy Request, instead queue tasks whose upstream nodes are complete.
    """
    update_erasure_tasks_with_access_data(session, privacy_request)
    ready_tasks: List[RequestTask] = (
        get_existing_ready_tasks(session, privacy_request, ActionType.erasure) or []
    )

    for task in ready_tasks:
        log_task_queued(task, "main runner")
        queue_request_task(task, privacy_request_proceed)
    return ready_tasks


@log_context
def run_consent_request(  # pylint: disable = too-many-arguments
    privacy_request: PrivacyRequest,
    graph: DatasetGraph,
    identity: Dict[str, Any],
    session: Session,
    privacy_request_proceed: bool = True,
) -> List[RequestTask]:
    """
    DSR 3.0: Build the "consent" graph, add its tasks to the database and queue the root task.

    If we are reprocessing a Privacy Request, instead queue tasks whose upstream nodes are complete.

    The graph built is very simple: there are no relationships between the nodes, every node has
    identity data input and every node outputs whether the consent request succeeded.

    The DatasetGraph passed in is expected to have one Node per Dataset.  That Node is expected to carry out requests
    for the Dataset as a whole.
    """

    if privacy_request.consent_tasks.count():
        ready_tasks: List[RequestTask] = get_existing_ready_tasks(
            session, privacy_request, ActionType.consent
        )
    else:
        logger.info("Building consent graph")
        traversal_nodes: Dict[CollectionAddress, TraversalNode] = {}
        # Unlike erasure and access graphs, we don't call traversal.traverse, but build a simpler
        # graph that just has one node per dataset
        for col_address, node in graph.nodes.items():
            traversal_node = TraversalNode(node)
            traversal_nodes[col_address] = traversal_node

        ready_tasks = persist_new_consent_request_tasks(
            session, privacy_request, traversal_nodes, identity, graph
        )

    for task in ready_tasks:
        log_task_queued(task, "main runner")
        queue_request_task(task, privacy_request_proceed)
    return ready_tasks


def get_existing_ready_tasks(
    session: Session, privacy_request: PrivacyRequest, action_type: ActionType
) -> List[RequestTask]:
    """
    Return existing RequestTasks if applicable in the event of reprocessing instead
    of creating new ones
    """
    ready: List[RequestTask] = []
    request_tasks: Query = privacy_request.get_tasks_by_action(action_type)
    if request_tasks.count():
        incomplete_tasks: Query = request_tasks.filter(
            RequestTask.status.notin_(COMPLETED_EXECUTION_LOG_STATUSES)
        )

        for task in incomplete_tasks:
            # Checks if both upstream tasks are complete and the task is not currently in-flight (if using workers)
            if task.can_queue_request_task(session, should_log=True):
                task.update_status(session, ExecutionLogStatus.pending)
                ready.append(task)
            elif task.status == ExecutionLogStatus.error:
                # Important to reset errored status to pending so it can be rerun
                task.update_status(session, ExecutionLogStatus.pending)

        if ready:
            logger.info(
                "Found existing {} task(s) ready to reprocess: {}.",
                action_type.value,
                [t.collection_address for t in ready],
            )
        return ready
    return ready



"""add manual task tables

Revision ID: 5efcdf18438e
Revises: c586a56c25e7
Create Date: 2025-06-04 17:24:00.300170

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql import func

# revision identifiers, used by Alembic.
revision = "5efcdf18438e"
down_revision = "c586a56c25e7"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "manual_task",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            nullable=False,
            server_default=func.now(),
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            nullable=False,
            server_default=func.now(),
        ),
        sa.Column(
            "task_type", sa.String(), nullable=False, server_default="privacy_request"
        ),
        sa.Column("parent_entity_id", sa.String(), nullable=False),
        sa.Column("parent_entity_type", sa.String(), nullable=False),
        sa.Column("due_date", sa.DateTime(timezone=True), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "parent_entity_id",
            "parent_entity_type",
            name="uq_manual_task_parent_entity",
        ),
    )

    op.create_table(
        "manual_task_reference",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            nullable=False,
            server_default=func.now(),
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            nullable=False,
            server_default=func.now(),
        ),
        sa.Column("task_id", sa.String(), nullable=False),
        sa.Column("reference_id", sa.String(), nullable=False),
        sa.Column("reference_type", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["task_id"],
            ["manual_task.id"],
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id"),
    )

    op.create_table(
        "manual_task_log",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            nullable=False,
            server_default=func.now(),
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            nullable=False,
            server_default=func.now(),
        ),
        sa.Column("task_id", sa.String(), nullable=False),
        sa.Column("config_id", sa.String(), nullable=True),
        sa.Column("instance_id", sa.String(), nullable=True),
        sa.Column("status", sa.String(), nullable=False),
        sa.Column("message", sa.String(), nullable=True),
        sa.Column("details", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.ForeignKeyConstraint(
            ["task_id"],
            ["manual_task.id"],
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id"),
    )

    # Create indexes for manual_task
    op.create_index("ix_manual_task_task_type", "manual_task", ["task_type"])
    op.create_index(
        "ix_manual_task_parent_entity",
        "manual_task",
        ["parent_entity_type", "parent_entity_id"],
    )
    op.create_index("ix_manual_task_due_date", "manual_task", ["due_date"])

    # Create indexes for manual_task_reference
    op.create_index(
        "ix_manual_task_reference_task_id", "manual_task_reference", ["task_id"]
    )
    op.create_index(
        "ix_manual_task_reference_reference",
        "manual_task_reference",
        ["reference_id", "reference_type"],
    )

    # Create indexes for manual_task_log
    op.create_index("ix_manual_task_log_task_id", "manual_task_log", ["task_id"])
    op.create_index("ix_manual_task_log_config_id", "manual_task_log", ["config_id"])
    op.create_index(
        "ix_manual_task_log_instance_id", "manual_task_log", ["instance_id"]
    )
    op.create_index("ix_manual_task_log_status", "manual_task_log", ["status"])
    op.create_index("ix_manual_task_log_created_at", "manual_task_log", ["created_at"])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    # Drop indexes first
    op.drop_index("ix_manual_task_log_created_at", table_name="manual_task_log")
    op.drop_index("ix_manual_task_log_status", table_name="manual_task_log")
    op.drop_index("ix_manual_task_log_instance_id", table_name="manual_task_log")
    op.drop_index("ix_manual_task_log_config_id", table_name="manual_task_log")
    op.drop_index("ix_manual_task_log_task_id", table_name="manual_task_log")
    op.drop_index(
        "ix_manual_task_reference_reference", table_name="manual_task_reference"
    )
    op.drop_index(
        "ix_manual_task_reference_task_id", table_name="manual_task_reference"
    )
    op.drop_index("ix_manual_task_due_date", table_name="manual_task")
    op.drop_index("ix_manual_task_parent_entity", table_name="manual_task")
    op.drop_index("ix_manual_task_task_type", table_name="manual_task")

    # Then drop tables
    op.drop_table("manual_task_log")
    op.drop_table("manual_task_reference")
    op.drop_table("manual_task")
    # ### end Alembic commands ###



"""add_active_column_taxonomy_elements

Revision ID: 5abb65a8cb91
Revises: 7c562441c589
Create Date: 2023-07-14 11:08:55.169966

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "5abb65a8cb91"
down_revision = "7c562441c589"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "ctl_data_categories",
        sa.Column("active", sa.BOOLEAN(), nullable=False, server_default="t"),
    )
    op.add_column(
        "ctl_data_subjects",
        sa.Column("active", sa.BOOLEAN(), nullable=False, server_default="t"),
    )
    op.add_column(
        "ctl_data_uses",
        sa.Column("active", sa.BOOLEAN(), nullable=False, server_default="t"),
    )
    op.add_column(
        "ctl_data_qualifiers",
        sa.Column("active", sa.BOOLEAN(), nullable=False, server_default="t"),
    )


def downgrade():
    op.drop_column("ctl_data_uses", "active")
    op.drop_column("ctl_data_subjects", "active")
    op.drop_column("ctl_data_categories", "active")
    op.drop_column("ctl_data_qualifiers", "active")



import os
from enum import Enum as EnumType
from io import BytesIO
from typing import IO, TYPE_CHECKING, Any, Tuple

from fideslang.validation import AnyHttpUrlString
from loguru import logger as log
from sqlalchemy import Column, DateTime
from sqlalchemy import Enum as EnumColumn
from sqlalchemy import ForeignKey, Index, String, UniqueConstraint, func, orm
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session, relationship

from fides.api.db.base_class import Base
from fides.api.schemas.storage.storage import StorageDetails, StorageType
from fides.api.service.storage.gcs import get_gcs_client
from fides.api.service.storage.s3 import (
    generic_delete_from_s3,
    generic_retrieve_from_s3,
    generic_upload_to_s3,
)
from fides.api.service.storage.util import AllowedFileType, get_local_filename

if TYPE_CHECKING:
    from fides.api.models.comment import Comment
    from fides.api.models.fides_user import FidesUser
    from fides.api.models.privacy_request import PrivacyRequest
    from fides.api.models.storage import StorageConfig


# This is 7 days in seconds and is currently the max allowed
# configurable expiration time for presigned URLs for both s3 and gcs.
MAX_TTL_SECONDS = 604800


class AttachmentType(str, EnumType):
    """
    Enum for attachment types. Indicates attachment usage.
    """

    internal_use_only = "internal_use_only"
    include_with_access_package = "include_with_access_package"


class AttachmentReferenceType(str, EnumType):
    """
    Enum for attachment reference types. Indicates where attachment is referenced.
    """

    access_manual_webhook = "access_manual_webhook"
    erasure_manual_webhook = "erasure_manual_webhook"
    privacy_request = "privacy_request"
    comment = "comment"
    manual_task_submission = "manual_task_submission"


class AttachmentReference(Base):
    """
    Stores information about an Attachment and any other element which may reference that attachment.
    """

    @declared_attr
    def __tablename__(cls) -> str:
        """Overriding base class method to set the table name."""
        return "attachment_reference"

    created_at = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )

    attachment_id = Column(
        String,
        ForeignKey("attachment.id", name="attachment_reference_attachment_id_fkey"),
        nullable=False,
    )
    reference_id = Column(String, nullable=False)
    reference_type = Column(EnumColumn(AttachmentReferenceType), nullable=False)

    __table_args__ = (
        UniqueConstraint(
            "attachment_id", "reference_id", name="_attachment_reference_uc"
        ),
        Index("ix_attachment_reference_reference_id", "reference_id"),
        Index("ix_attachment_reference_reference_type", "reference_type"),
    )

    # Relationships
    attachment = relationship("Attachment", back_populates="references")

    @classmethod
    def create(
        cls, db: Session, *, data: dict[str, Any], check_name: bool = False
    ) -> "AttachmentReference":
        """Creates a new attachment reference record in the database."""
        return super().create(db=db, data=data, check_name=check_name)


class Attachment(Base):
    """
    Stores information about an Attachment.
    """

    created_at = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )

    user_id = Column(
        String, ForeignKey("fidesuser.id", ondelete="SET NULL"), nullable=True
    )
    file_name = Column(String, nullable=False)
    attachment_type = Column(EnumColumn(AttachmentType), nullable=False)
    storage_key = Column(
        String, ForeignKey("storageconfig.key", ondelete="CASCADE"), nullable=False
    )

    user = relationship(
        "FidesUser",
        lazy="selectin",
        uselist=False,
    )

    references = relationship(
        "AttachmentReference",
        back_populates="attachment",
        cascade="all, delete-orphan",
        uselist=True,
        foreign_keys=[AttachmentReference.attachment_id],
        primaryjoin=lambda: Attachment.id
        == orm.foreign(AttachmentReference.attachment_id),
    )

    config = relationship(
        "StorageConfig",
        lazy="selectin",
        uselist=False,
    )

    @property
    def content_type(self) -> str:
        """Returns the content type of the attachment."""
        return AllowedFileType[self.file_name.split(".")[-1]].value

    @property
    def file_key(self) -> str:
        """Returns the file key of the attachment."""
        return f"{self.id}/{self.file_name}"

    def upload(self, attachment: IO[bytes]) -> None:
        """Uploads an attachment to S3, GCS, or local storage."""
        if self.config.type == StorageType.s3:
            bucket_name = f"{self.config.details[StorageDetails.BUCKET.value]}"
            auth_method = self.config.details[StorageDetails.AUTH_METHOD.value]
            generic_upload_to_s3(
                storage_secrets=self.config.secrets,
                bucket_name=bucket_name,
                file_key=self.file_key,
                document=attachment,
                auth_method=auth_method,
            )
            log.info(f"Uploaded {self.file_name} to S3 bucket {bucket_name}/{self.id}")
            return

        if self.config.type == StorageType.gcs:
            bucket_name = f"{self.config.details[StorageDetails.BUCKET.value]}"
            auth_method = self.config.details[StorageDetails.AUTH_METHOD.value]
            storage_client = get_gcs_client(auth_method, self.config.secrets)
            bucket = storage_client.bucket(bucket_name)
            blob = bucket.blob(self.file_key)

            # Reset the file pointer to the beginning
            try:
                attachment.seek(0)
            except Exception as e:
                raise ValueError(f"Failed to reset file pointer for attachment: {e}")

            blob.upload_from_file(attachment)
            log.info(f"Uploaded {self.file_name} to GCS bucket {bucket_name}/{self.id}")
            return

        if self.config.type == StorageType.local:
            filename = get_local_filename(self.file_key)

            # Validate that attachment is a file-like object
            if not hasattr(attachment, "read"):
                raise TypeError(f"Expected a file-like object, got {type(attachment)}")

            # Reset the file pointer to the beginning
            try:
                attachment.seek(0)
            except Exception as e:
                raise ValueError(f"Failed to reset file pointer for attachment: {e}")

            # Create directory if it doesn't exist
            os.makedirs(os.path.dirname(filename), exist_ok=True)

            # Write the file in chunks to avoid loading the entire content into memory
            with open(filename, "wb") as file:
                for chunk in iter(
                    lambda: attachment.read(1024 * 1024), b""
                ):  # 1 MB chunks
                    if not isinstance(chunk, bytes):
                        raise TypeError(f"Expected bytes, got {type(chunk)}")
                    file.write(chunk)

            log.info(f"Uploaded {self.file_name} to local storage at {filename}")
            return

        raise ValueError(f"Unsupported storage type: {self.config.type}")

    def retrieve_attachment(
        self,
    ) -> Tuple[int, AnyHttpUrlString]:
        """
        Retrieves the size of the attachment and a presigned/signed URL.
        - For s3:
          - the size is retrieved from the s3 object metadata
          - returns presigned URL
        - For gcs:
          - the size is retrieved from the blob metadata
          - returns signed URL
        - For local:
          - the size is retrieved from the file size
          - returns the local file path
        """
        if self.config.type == StorageType.s3:
            bucket_name = f"{self.config.details[StorageDetails.BUCKET.value]}"
            auth_method = self.config.details[StorageDetails.AUTH_METHOD.value]
            size, url = generic_retrieve_from_s3(
                storage_secrets=self.config.secrets,
                bucket_name=bucket_name,
                file_key=self.file_key,
                auth_method=auth_method,
                get_content=False,
                ttl_seconds=MAX_TTL_SECONDS,
            )
            return size, url

        if self.config.type == StorageType.gcs:
            bucket_name = f"{self.config.details[StorageDetails.BUCKET.value]}"
            auth_method = self.config.details[StorageDetails.AUTH_METHOD.value]
            storage_client = get_gcs_client(auth_method, self.config.secrets)
            bucket = storage_client.bucket(bucket_name)
            blob = bucket.blob(self.file_key)

            # Ensure we have the blob metadata
            blob.reload()
            # Expiration is set to 7 days
            url = blob.generate_signed_url(
                version="v4",
                expiration=MAX_TTL_SECONDS,
                method="GET",
            )
            return blob.size, url

        if self.config.type == StorageType.local:
            filename = get_local_filename(self.file_key)
            size = os.path.getsize(filename)
            return size, filename

        raise ValueError(f"Unsupported storage type: {self.config.type}")

    def retrieve_attachment_content(
        self,
    ) -> Tuple[int, IO[bytes]]:
        """
        Retrieves the size of the attachment and its actual content.
        - For s3:
          - the size is retrieved from the s3 object metadata
          - returns the actual content
        - For gcs:
          - the size is retrieved from the blob metadata
          - returns the actual content
        - For local:
          - the size is retrieved from the file size
          - returns the actual content
        """
        if self.config.type == StorageType.s3:
            bucket_name = f"{self.config.details[StorageDetails.BUCKET.value]}"
            auth_method = self.config.details[StorageDetails.AUTH_METHOD.value]
            size, fileobj = generic_retrieve_from_s3(
                storage_secrets=self.config.secrets,
                bucket_name=bucket_name,
                file_key=self.file_key,
                auth_method=auth_method,
                get_content=True,
            )
            return size, fileobj

        if self.config.type == StorageType.gcs:
            bucket_name = f"{self.config.details[StorageDetails.BUCKET.value]}"
            auth_method = self.config.details[StorageDetails.AUTH_METHOD.value]
            storage_client = get_gcs_client(auth_method, self.config.secrets)
            bucket = storage_client.bucket(bucket_name)
            blob = bucket.blob(self.file_key)

            fileobj = BytesIO()
            blob.download_to_file(fileobj)
            fileobj.seek(0)  # Reset pointer to beginning after download
            return blob.size, fileobj

        if self.config.type == StorageType.local:
            filename = get_local_filename(self.file_key)
            size = os.path.getsize(filename)
            with open(filename, "rb") as fileobj:
                content = BytesIO(fileobj.read())
                content.seek(0)  # Reset pointer to beginning
                return size, content

        raise ValueError(f"Unsupported storage type: {self.config.type}")

    def delete_attachment_from_storage(self) -> None:
        """Deletes an attachment from S3, GCS, or local storage."""
        if self.config.type == StorageType.s3:
            bucket_name = f"{self.config.details[StorageDetails.BUCKET.value]}"
            auth_method = self.config.details[StorageDetails.AUTH_METHOD.value]
            generic_delete_from_s3(
                storage_secrets=self.config.secrets,
                bucket_name=bucket_name,
                file_key=self.file_key,
                auth_method=auth_method,
            )
            return

        if self.config.type == StorageType.gcs:
            bucket_name = f"{self.config.details[StorageDetails.BUCKET.value]}"
            auth_method = self.config.details[StorageDetails.AUTH_METHOD.value]
            storage_client = get_gcs_client(auth_method, self.config.secrets)
            bucket = storage_client.bucket(bucket_name)

            # List and delete all blobs in the folder
            prefix = f"{self.id}/"
            blobs = bucket.list_blobs(prefix=prefix)
            for blob in blobs:
                blob.delete()
            return

        if self.config.type == StorageType.local:
            folder_path = os.path.dirname(get_local_filename(self.file_key))
            if os.path.exists(folder_path):
                import shutil

                shutil.rmtree(folder_path)
            return

        raise ValueError(f"Unsupported storage type: {self.config.type}")

    @classmethod
    def create_and_upload(
        cls,
        db: Session,
        *,
        data: dict[str, Any],
        attachment_file: IO[bytes],
        check_name: bool = False,
    ) -> "Attachment":
        """Creates a new attachment record in the database and uploads the attachment via the upload method."""
        attachment_model = super().create(db=db, data=data, check_name=check_name)

        try:
            attachment_model.upload(attachment_file)
            return attachment_model
        except Exception as e:
            log.error(f"Failed to upload attachment: {e}")
            attachment_model.delete(db)
            raise e

    @classmethod
    def create(
        cls,
        db: Session,
        *,
        data: dict[str, Any],
        check_name: bool = False,
    ) -> "Attachment":
        """Raises Error, provides information for user to create with upload instead."""
        raise NotImplementedError("Please use create_and_upload method for Attachment")

    def delete(self, db: Session) -> None:
        """Deletes an attachment record from the database and deletes the attachment from S3."""
        self.delete_attachment_from_storage()

        # Delete all references to the attachment
        for reference in self.references:
            reference.delete(db)
        super().delete(db=db)

    @staticmethod
    def delete_attachments_for_reference_and_type(
        db: Session, reference_id: str, reference_type: AttachmentReferenceType
    ) -> None:
        """
        Deletes attachments associated with a given reference_id and reference_type.
        Deletes all references to the attachments.

        Args:
            db: Database session
            reference_id: ID of the reference
            reference_type: Type of the reference

        Examples:

        - Delete all attachments associated with a comment.
           ``Attachment.delete_attachments_for_reference_and_type(
               db, comment.id, AttachmentReferenceType.comment
           )``
        - Delete all attachments associated with a privacy request.
           ``Attachment.delete_attachments_for_reference_and_type(
               db, privacy_request.id, AttachmentReferenceType.privacy_request
            )``
        """
        # Query attachments explicitly to avoid lazy loading
        attachments = (
            db.query(Attachment)
            .join(AttachmentReference)
            .filter(
                AttachmentReference.reference_id == reference_id,
                AttachmentReference.reference_type == reference_type,
            )
            .all()
        )

        for attachment in attachments:
            attachment.delete(db)



from fides.api.models.privacy_notice import UserConsentPreference
from fides.api.schemas.base_class import FidesSchema
from fides.api.schemas.privacy_notice import PrivacyNoticeHistorySchema


class MinimalPrivacyPreferenceHistorySchema(FidesSchema):
    """Minimal privacy preference history schema for building consent emails"""

    preference: UserConsentPreference
    privacy_notice_history: PrivacyNoticeHistorySchema



"""
Utilities for storing and accessing request-scoped context that must be
accessible across the entire application stack (endpoints -> services ->
helpers -> decorators) without having to thread additional parameters through
all call-sites.

Currently we only capture the authenticated `user_id` but additional fields
(e.g. correlation_id, locale, feature_flags) can be added in the future by
expanding the `RequestContext` dataclass.

A `contextvars.ContextVar` is used instead of a module-level global to ensure
that values are local to the current execution context (async task, thread or
Celery worker) and therefore safe for concurrent workloads.
"""

from __future__ import annotations

import contextvars
from dataclasses import dataclass
from typing import Any, Optional

__all__ = [
    "get_user_id",
    "set_user_id",
    "reset_request_context",
]


@dataclass
class RequestContext:
    user_id: Optional[str] = None


# A single ContextVar holding the current request context.
_ctx: contextvars.ContextVar[RequestContext] = contextvars.ContextVar(
    "request_context",
    default=RequestContext(),
)


def get_request_context() -> RequestContext:
    """Return the current `RequestContext` for this request.

    - The context is populated during authentication by `extract_token_and_load_client()`,
      which calls `fides.api.request_context.set_request_context` with
      the authenticated `user_id`.
    - A `ContextVar` keeps the data isolated per request. Each coroutine
      or thread sees its own copy and it is discarded at the end of the request.
    - The returned object is the live context; treat it as read-only and use
      `set_request_context` to mutate values.
    """
    return _ctx.get()


def set_request_context(**kwargs: Any) -> None:
    """Mutate the current context in place using `key=value` pairs."""
    ctx = _ctx.get()
    for key, value in kwargs.items():
        if hasattr(ctx, key):
            setattr(ctx, key, value)


def reset_request_context() -> None:
    """Remove all context, mainly for test clean-up."""
    _ctx.set(RequestContext())


def get_user_id() -> Optional[str]:
    """Return the user_id from the current request context."""
    ctx = get_request_context()
    return ctx.user_id


def set_user_id(user_id: str) -> None:
    """Set the user_id in the current request context."""
    set_request_context(user_id=user_id)



from datetime import datetime, timezone
from enum import Enum
from typing import TYPE_CHECKING, Annotated, Any, Optional, cast

from pydantic import ConfigDict, Field
from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Index,
    Integer,
    String,
    UniqueConstraint,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session, relationship
from sqlalchemy.sql import func

from fides.api.db.base_class import Base, FidesBase
from fides.api.db.util import EnumColumn
from fides.api.request_context import get_user_id
from fides.api.schemas.base_class import FidesSchema

if TYPE_CHECKING:
    from fides.api.models.attachment import Attachment
    from fides.api.models.fides_user import FidesUser
    from fides.api.models.manual_task.conditional_dependency import (
        ManualTaskConditionalDependency,
    )

# ------------------------------------------------------------
# Enums
# ------------------------------------------------------------


class ManualTaskExecutionTiming(str, Enum):
    """Enum for when a manual task should be executed in the privacy request DAG."""

    pre_execution = "pre_execution"  # Execute before the main DAG
    post_execution = "post_execution"  # Execute after the main DAG
    parallel = "parallel"  # Execute in parallel with the main DAG


class ManualTaskType(str, Enum):
    """Enum for manual task types."""

    privacy_request = "privacy_request"
    # Add more task types as needed


class ManualTaskParentEntityType(str, Enum):
    """Enum for manual task parent entity types."""

    connection_config = (
        "connection_config"  # used for access and erasure privacy requests
    )
    # Add more parent entity types as needed


class ManualTaskEntityType(str, Enum):
    """Enum for manual task entity types."""

    privacy_request = "privacy_request"
    # Add more entity types as needed


class ManualTaskReferenceType(str, Enum):
    """Enum for manual task reference types."""

    privacy_request = "privacy_request"
    connection_config = "connection_config"
    manual_task_config = "manual_task_config"
    assigned_user = "assigned_user"  # Reference to the user assigned to the task
    # Add more reference types as needed


class ManualTaskLogStatus(str, Enum):
    """Enum for manual task log status."""

    created = "created"
    updated = "updated"
    in_progress = "in_progress"
    complete = "complete"
    error = "error"
    retrying = "retrying"
    paused = "paused"
    awaiting_input = "awaiting_input"


class ManualTaskConfigurationType(str, Enum):
    """Enum for manual task configuration types."""

    access_privacy_request = "access_privacy_request"
    erasure_privacy_request = "erasure_privacy_request"
    # Add more configuration types as needed


class ManualTaskFieldType(str, Enum):
    """Enum for manual task field types."""

    text = "text"  # Key-value pairs
    checkbox = "checkbox"  # Boolean value
    attachment = "attachment"  # File upload
    # Add more field types as needed


class StatusType(str, Enum):
    """Enum for manual task status."""

    pending = "pending"
    in_progress = "in_progress"
    completed = "completed"
    failed = "failed"

    @classmethod
    def get_valid_transitions(cls, current_status: "StatusType") -> list["StatusType"]:
        """Get valid transitions from the current status."""
        transitions = {
            cls.pending: [cls.in_progress, cls.failed, cls.completed],
            cls.in_progress: [cls.completed, cls.failed],
            cls.completed: [],
            cls.failed: [cls.pending, cls.in_progress],
        }
        return transitions.get(current_status, [])


# ------------------------------------------------------------
# Schemas
# ------------------------------------------------------------


class ManualTaskFieldMetadata(FidesSchema):
    """Base schema for manual task field metadata."""

    label: Annotated[str, Field(description="Display label for the field")]
    required: Annotated[
        bool, Field(default=False, description="Whether the field is required")
    ]
    help_text: Annotated[
        Optional[str],
        Field(default=None, description="Help text to display with the field"),
    ]
    data_uses: Annotated[
        Optional[list[str]],
        Field(
            default=None,
            description="List of data uses associated with this field",
        ),
    ]

    model_config = ConfigDict(extra="allow")


# ------------------------------------------------------------
# Models
# ------------------------------------------------------------


class ManualTask(Base):
    """Model for storing manual tasks.

    This model can be used for both privacy request tasks and general tasks.
    For privacy requests, it replaces the functionality of manual webhooks.
    For other use cases, it provides a flexible task management system.

    There can only be one ManualTask per parent entity.
    You can create multiple Configs for the same ManualTask.
    """

    @declared_attr
    def __tablename__(cls) -> str:
        """Overriding base class method to set the table name."""
        return "manual_task"

    # redefined here because there's a minor, unintended discrepancy between
    # this `id` field and that of the `Base` class, which explicitly sets `index=True`.
    # TODO: we likely should _not_ be setting `index=True` on the `id`
    # attribute of the `Base` class, as `primary_key=True` already specifies a
    # primary key constraint, which will implicitly create an index for the field.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)
    created_at = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )

    # Database columns
    task_type = Column(
        EnumColumn(ManualTaskType),
        nullable=False,
        default=ManualTaskType.privacy_request,
    )
    parent_entity_id = Column(String, nullable=False)
    parent_entity_type = Column(
        EnumColumn(ManualTaskParentEntityType),
        nullable=False,
        default=ManualTaskParentEntityType.connection_config,
    )

    __table_args__ = (
        UniqueConstraint(
            "parent_entity_id",
            "parent_entity_type",
            name="uq_manual_task_parent_entity",
        ),
        Index("ix_manual_task_parent_entity", "parent_entity_type", "parent_entity_id"),
        Index("ix_manual_task_task_type", "task_type"),
    )

    # Relationships
    references = relationship(
        "ManualTaskReference",
        back_populates="task",
        uselist=True,
        cascade="all, delete-orphan",
    )
    logs = relationship(
        "ManualTaskLog",
        back_populates="task",
        primaryjoin="and_(ManualTask.id == ManualTaskLog.task_id)",
        order_by="ManualTaskLog.created_at",
    )
    configs = relationship(
        "ManualTaskConfig",
        back_populates="task",
        uselist=True,
        viewonly=True,  # No cascade delete - configs are versioned
    )
    instances = relationship(
        "ManualTaskInstance",
        back_populates="task",
        uselist=True,
        viewonly=True,  # No cascade delete - instances contain historical data
    )
    submissions = relationship(
        "ManualTaskSubmission",
        back_populates="task",
        uselist=True,
        viewonly=True,  # No cascade delete - submissions are historical data
    )

    conditional_dependencies = relationship(
        "ManualTaskConditionalDependency",
        back_populates="task",
        uselist=True,
        cascade="all, delete-orphan",
    )

    # Properties
    @property
    def assigned_users(self) -> list[str]:
        """Get all users assigned to this task."""
        if not self.references:
            return []
        return [
            ref.reference_id
            for ref in self.references
            if ref.reference_type == ManualTaskReferenceType.assigned_user
        ]

    # CRUD Operations
    @classmethod
    def create(
        cls, db: Session, *, data: dict[str, Any], check_name: bool = True
    ) -> "ManualTask":
        """Create a new manual task."""
        task = super().create(db=db, data=data, check_name=check_name)
        ManualTaskLog.create_log(
            db=db,
            task_id=task.id,
            status=ManualTaskLogStatus.created,
            message=f"Created manual task for {data['task_type']}",
        )
        return task


class ManualTaskInstance(Base):
    """Model for tracking task status per entity instance."""

    @declared_attr
    def __tablename__(cls) -> str:
        """Overriding base class method to set the table name."""
        return "manual_task_instance"

    # redefined here because there's a minor, unintended discrepancy between
    # this `id` field and that of the `Base` class, which explicitly sets `index=True`.
    # TODO: we likely should _not_ be setting `index=True` on the `id`
    # attribute of the `Base` class, as `primary_key=True` already specifies a
    # primary key constraint, which will implicitly create an index for the field.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)
    created_at = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )

    # Database columns
    task_id = Column(
        String,
        ForeignKey(
            "manual_task.id",
            name="manual_task_instance_task_id_fkey",
            ondelete="SET NULL",
        ),
        nullable=True,
    )
    config_id = Column(
        String,
        ForeignKey(
            "manual_task_config.id",
            name="manual_task_instance_config_id_fkey",
            ondelete="RESTRICT",
        ),
        nullable=False,
    )
    # entity id is the entity that the instance relates to
    # (e.g. a privacy request is an entity that has its own manual task instance)
    entity_id = Column(String, nullable=False)
    entity_type = Column(EnumColumn(ManualTaskEntityType), nullable=False)
    # ignore[assignment] because the mypy and sqlalchemy types mismatch
    # upgrading to 2.0 allows mapping which provides better type safety visibility.
    status = Column(EnumColumn(StatusType), nullable=False, default=StatusType.pending)  # type: ignore[assignment]
    completed_at = Column(DateTime, nullable=True)  # type: ignore[assignment]
    completed_by_id = Column(String, nullable=True)  # type: ignore[assignment]
    due_date = Column(DateTime, nullable=True)

    __table_args__ = (
        Index("ix_manual_task_instance_completed_at", "completed_at"),
        Index("ix_manual_task_instance_config_id", "config_id"),
        Index("ix_manual_task_instance_entity", "entity_type", "entity_id"),
        Index("ix_manual_task_instance_entity_id", "entity_id"),
        Index("ix_manual_task_instance_entity_type", "entity_type"),
        Index("ix_manual_task_instance_status", "status"),
        Index("ix_manual_task_instance_task_id", "task_id"),
    )

    # Relationships
    task = relationship("ManualTask", back_populates="instances")
    config = relationship("ManualTaskConfig", back_populates="instances")
    submissions = relationship(
        "ManualTaskSubmission",
        back_populates="instance",
        uselist=True,
    )
    logs = relationship(
        "ManualTaskLog",
        back_populates="instance",
        primaryjoin="ManualTaskInstance.id == ManualTaskLog.instance_id",
        cascade="all, delete-orphan",
        order_by="ManualTaskLog.created_at",
        uselist=True,
    )
    attachments = relationship(
        "Attachment",
        secondary="attachment_reference",
        primaryjoin="and_(ManualTaskInstance.id == ManualTaskSubmission.instance_id, "
        "ManualTaskSubmission.id == AttachmentReference.reference_id, "
        "AttachmentReference.reference_type == 'manual_task_submission')",
        secondaryjoin="Attachment.id == AttachmentReference.attachment_id",
        order_by="Attachment.created_at",
        viewonly=True,
        uselist=True,
    )

    @property
    def required_fields(self) -> list["ManualTaskConfigField"]:
        """Get all required fields."""
        return [
            field
            for field in self.config.field_definitions
            if field.field_metadata.get("required", False)
        ]

    @property
    def incomplete_fields(self) -> list["ManualTaskConfigField"]:
        """Get all fields that haven't been completed yet.
        A field is considered incomplete if:
        1. It's required and has no submission
        Returns:
            list[ManualTaskConfigField]: List of incomplete fields
        """
        return [
            field
            for field in self.required_fields
            if not self.get_submission_for_field(field.id)
        ]

    @property
    def completed_fields(self) -> list["ManualTaskConfigField"]:
        """Get all fields that have been completed."""
        return [
            field
            for field in self.config.field_definitions
            if field.field_metadata.get("required", False)
            and self.get_submission_for_field(field.id)
        ]

    def get_submission_for_field(
        self, field_id: str
    ) -> Optional["ManualTaskSubmission"]:
        """Get the submission for a specific field.

        Args:
            field_id: The ID of the field to get the submission for

        Returns:
            Optional[ManualTaskSubmission]: The submission for the field, or None if no submission exists
        """
        return next(
            (
                submission
                for submission in self.submissions
                if submission.field_id == field_id
            ),
            None,
        )

    def update_status(
        self, db: Session, new_status: StatusType, user_id: Optional[str] = None
    ) -> None:
        """Update the status with completion handling."""

        if new_status == StatusType.completed:
            self.completed_at = datetime.now(timezone.utc)
            self.completed_by_id = user_id
        elif new_status == StatusType.pending:
            # Reset completion fields if going back to pending
            self.completed_at = None
            self.completed_by_id = None

        self.status = new_status
        db.add(self)
        db.commit()

    def mark_completed(self, db: Session, user_id: str) -> None:
        """Mark as completed."""
        self.update_status(db, StatusType.completed, user_id)

    def delete(self, db: Session) -> None:
        """Delete the instance and all associated submissions with proper attachment cleanup."""
        # Delete submissions one by one to ensure proper attachment cleanup
        for submission in self.submissions:
            submission.delete(db)
        # Delete the instance itself
        db.delete(self)


class ManualTaskReference(Base):
    """Join table to associate manual tasks with multiple references.

    A single task may have many references including privacy requests, configurations, and assigned users.
    """

    @declared_attr
    def __tablename__(cls) -> str:
        """Overriding base class method to set the table name."""
        return "manual_task_reference"

    # redefined here because there's a minor, unintended discrepancy between
    # this `id` field and that of the `Base` class, which explicitly sets `index=True`.
    # TODO: we likely should _not_ be setting `index=True` on the `id`
    # attribute of the `Base` class, as `primary_key=True` already specifies a
    # primary key constraint, which will implicitly create an index for the field.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)
    created_at = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )

    # Database columns
    task_id = Column(
        String, ForeignKey("manual_task.id", ondelete="CASCADE"), nullable=False
    )
    reference_id = Column(String, nullable=False)
    reference_type = Column(EnumColumn(ManualTaskReferenceType), nullable=False)

    __table_args__ = (
        Index("ix_manual_task_reference_reference", "reference_id", "reference_type"),
        Index("ix_manual_task_reference_task_id", "task_id"),
    )

    # Relationships
    task = relationship("ManualTask", back_populates="references")


class ManualTaskConfig(Base):
    """Model for storing manual task configurations.
    A single configuration may have many fields of different types.
    """

    @declared_attr
    def __tablename__(cls) -> str:
        return "manual_task_config"

    # redefined here because there's a minor, unintended discrepancy between
    # this `id` field and that of the `Base` class, which explicitly sets `index=True`.
    # TODO: we likely should _not_ be setting `index=True` on the `id`
    # attribute of the `Base` class, as `primary_key=True` already specifies a
    # primary key constraint, which will implicitly create an index for the field.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)
    created_at = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )

    task_id = Column(
        String,
        ForeignKey(
            "manual_task.id",
            name="manual_task_config_task_id_fkey",
            ondelete="SET NULL",
        ),
        nullable=True,
    )
    config_type = Column(EnumColumn(ManualTaskConfigurationType), nullable=False)
    version = Column(Integer, nullable=False, default=1)
    is_current = Column(Boolean, nullable=False, default=True)
    execution_timing = Column(
        EnumColumn(ManualTaskExecutionTiming),
        nullable=False,
        default=ManualTaskExecutionTiming.pre_execution,
    )

    __table_args__ = (
        Index("ix_manual_task_config_config_type", "config_type"),
        Index("ix_manual_task_config_task_id", "task_id"),
    )

    # Relationships
    task = relationship("ManualTask", back_populates="configs")
    instances = relationship(
        "ManualTaskInstance",
        back_populates="config",
        uselist=True,
        viewonly=True,
    )
    submissions = relationship(
        "ManualTaskSubmission",
        back_populates="config",
        uselist=True,
        viewonly=True,
    )
    field_definitions = relationship(
        "ManualTaskConfigField",
        back_populates="config",
        cascade="all, delete-orphan",
        uselist=True,
    )
    logs = relationship(
        "ManualTaskLog",
        back_populates="config",
        primaryjoin="ManualTaskConfig.id == ManualTaskLog.config_id",
        cascade="all, delete-orphan",
    )

    @classmethod
    def create(
        cls, db: Session, *, data: dict[str, Any], check_name: bool = True
    ) -> "ManualTaskConfig":
        """Create a new manual task configuration."""
        # Validate config_type
        try:
            ManualTaskConfigurationType(data["config_type"])
        except ValueError:
            raise ValueError(f"Invalid config type: {data['config_type']}")

        config = super().create(db=db, data=data, check_name=check_name)

        # Log the config creation as a task-level log
        ManualTaskLog.create_log(
            db=db,
            task_id=data["task_id"],
            config_id=config.id,
            status=ManualTaskLogStatus.created,
            message=f"Created manual task configuration for {data['config_type']}",
            details={
                "config_type": data["config_type"],
            },
        )
        return config

    def get_field(self, field_key: str) -> Optional["ManualTaskConfigField"]:
        """Get a field by its key."""
        for field in self.field_definitions:
            if field.field_key == field_key:
                return field
        return None


class ManualTaskConfigField(Base):
    """Model for storing fields associated with each config."""

    @declared_attr
    def __tablename__(cls) -> str:
        return "manual_task_config_field"

    # redefined here because there's a minor, unintended discrepancy between
    # this `id` field and that of the `Base` class, which explicitly sets `index=True`.
    # TODO: we likely should _not_ be setting `index=True` on the `id`
    # attribute of the `Base` class, as `primary_key=True` already specifies a
    # primary key constraint, which will implicitly create an index for the field.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)
    created_at = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )

    task_id = Column(
        String,
        ForeignKey(
            "manual_task.id",
            name="manual_task_config_field_task_id_fkey",
            ondelete="SET NULL",
        ),
        nullable=True,
    )
    config_id = Column(
        String,
        ForeignKey(
            "manual_task_config.id",
            name="manual_task_config_field_config_id_fkey",
            ondelete="CASCADE",
        ),
        nullable=False,
    )
    field_key = Column(String, nullable=False)
    field_type = Column(
        EnumColumn(ManualTaskFieldType), nullable=False
    )  # Using ManualTaskFieldType
    field_metadata: dict[str, Any] = cast(
        dict[str, Any], Column(JSONB, nullable=False, default={})
    )

    __table_args__ = (
        UniqueConstraint("config_id", "field_key", name="unique_field_key_per_config"),
        Index("ix_manual_task_config_field_config_id", "config_id"),
        Index("ix_manual_task_config_field_field_key", "field_key"),
        Index("ix_manual_task_config_field_task_id", "task_id"),
    )

    # Relationships
    config = relationship("ManualTaskConfig", back_populates="field_definitions")
    submissions = relationship(
        "ManualTaskSubmission",
        back_populates="field",
        uselist=True,
    )

    @property
    def field_metadata_model(self) -> ManualTaskFieldMetadata:
        """Get the field metadata as a Pydantic model."""
        assert isinstance(
            self.field_metadata, dict
        ), "field_metadata must be a dictionary"
        return ManualTaskFieldMetadata.model_validate(self.field_metadata)

    @classmethod
    def create(
        cls, db: Session, *, data: dict[str, Any], check_name: bool = True
    ) -> "ManualTaskConfigField":
        """Create a new manual task config field."""
        # Get the config to access its task_id and check if it exists
        config = (
            db.query(ManualTaskConfig)
            .filter(ManualTaskConfig.id == data["config_id"])
            .first()
        )
        if not config:
            raise ValueError(f"Config with id {data['config_id']} not found")

        # Create the field and let SQLAlchemy complex type validation handled in service.
        field = super().create(db=db, data=data, check_name=check_name)

        # Create a log entry
        if config.task_id:
            ManualTaskLog.create_log(
                db=db,
                task_id=config.task_id,
                config_id=data["config_id"],
                status=ManualTaskLogStatus.created,
                message=f"Created manual task config field for {data['field_key']}",
            )
        return field


class ManualTaskSubmission(Base):
    """Model for storing user submissions.
    Each submission represents data for a single field.
    """

    @declared_attr
    def __tablename__(cls) -> str:
        """Overriding base class method to set the table name."""
        return "manual_task_submission"

    # redefined here because there's a minor, unintended discrepancy between
    # this `id` field and that of the `Base` class, which explicitly sets `index=True`.
    # TODO: we likely should _not_ be setting `index=True` on the `id`
    # attribute of the `Base` class, as `primary_key=True` already specifies a
    # primary key constraint, which will implicitly create an index for the field.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)
    created_at = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )

    # Database columns
    task_id = Column(
        String,
        ForeignKey(
            "manual_task.id",
            name="manual_task_submission_task_id_fkey",
            ondelete="SET NULL",
        ),
        nullable=True,
    )
    config_id = Column(
        String,
        ForeignKey(
            "manual_task_config.id",
            name="manual_task_submission_config_id_fkey",
            ondelete="RESTRICT",
        ),
        nullable=False,
    )
    field_id = Column(
        String,
        ForeignKey(
            "manual_task_config_field.id",
            name="manual_task_submission_field_id_fkey",
            ondelete="RESTRICT",
        ),
        nullable=False,
    )
    instance_id = Column(
        String,
        ForeignKey(
            "manual_task_instance.id",
            name="manual_task_submission_instance_id_fkey",
            ondelete="CASCADE",
        ),
        nullable=False,
    )
    submitted_by = Column(
        String,
        ForeignKey(
            "fidesuser.id",
            name="manual_task_submission_submitted_by_fkey",
            ondelete="SET NULL",
        ),
        nullable=True,
    )
    submitted_at = Column(DateTime, default=datetime.now(timezone.utc), nullable=False)
    data = Column(JSONB, nullable=False)

    __table_args__ = (
        Index("ix_manual_task_submission_config_id", "config_id"),
        Index("ix_manual_task_submission_field_id", "field_id"),
        Index("ix_manual_task_submission_instance_field", "instance_id", "field_id"),
        Index("ix_manual_task_submission_instance_id", "instance_id"),
        Index("ix_manual_task_submission_submitted_at", "submitted_at"),
        Index("ix_manual_task_submission_submitted_by", "submitted_by"),
        Index("ix_manual_task_submission_task_id", "task_id"),
    )

    # Relationships
    task = relationship("ManualTask", back_populates="submissions")
    config = relationship("ManualTaskConfig", back_populates="submissions")
    field = relationship("ManualTaskConfigField", back_populates="submissions")
    instance = relationship(
        "ManualTaskInstance",
        back_populates="submissions",
    )
    attachments = relationship(
        "Attachment",
        secondary="attachment_reference",
        primaryjoin="and_(ManualTaskSubmission.id == AttachmentReference.reference_id, "
        "AttachmentReference.reference_type == 'manual_task_submission')",
        secondaryjoin="Attachment.id == AttachmentReference.attachment_id",
        order_by="Attachment.created_at",
        uselist=True,
        viewonly=True,  # Make read-only to avoid overlapping relationship warnings
    )

    user = relationship(
        "FidesUser",
        primaryjoin="FidesUser.id == ManualTaskSubmission.submitted_by",
        viewonly=True,
    )

    def delete(self, db: Session) -> None:
        """Delete the submission and all associated attachments."""
        from fides.api.models.attachment import Attachment, AttachmentReferenceType

        # Delete attachments associated with this submission
        Attachment.delete_attachments_for_reference_and_type(
            db, self.id, AttachmentReferenceType.manual_task_submission
        )
        # Delete the submission itself
        db.delete(self)

    @staticmethod
    def delete_submissions_for_instance(db: Session, instance_id: str) -> None:
        """
        Deletes submissions associated with a given instance_id.
        Properly handles attachment cleanup for each submission.

        Args:
            db: Database session.
            instance_id: The instance id to delete submissions for.

        Example:
          - Delete all submissions associated with a manual task instance.
            ``ManualTaskSubmission.delete_submissions_for_instance(
                db, instance.id
            )``
        """
        # Query submissions explicitly to avoid lazy loading
        submissions = (
            db.query(ManualTaskSubmission)
            .filter(ManualTaskSubmission.instance_id == instance_id)
            .all()
        )

        for submission in submissions:
            submission.delete(db)


class ManualTaskLog(Base):
    """Model for storing manual task execution logs."""

    @declared_attr
    def __tablename__(cls) -> str:
        """Overriding base class method to set the table name."""
        return "manual_task_log"

    # redefined here because there's a minor, unintended discrepancy between
    # this `id` field and that of the `Base` class, which explicitly sets `index=True`.
    # TODO: we likely should _not_ be setting `index=True` on the `id`
    # attribute of the `Base` class, as `primary_key=True` already specifies a
    # primary key constraint, which will implicitly create an index for the field.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)
    created_at = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )

    task_id = Column(
        String, ForeignKey("manual_task.id", ondelete="SET NULL"), nullable=True
    )
    config_id = Column(
        String, ForeignKey("manual_task_config.id", ondelete="CASCADE"), nullable=True
    )
    instance_id = Column(
        String,
        ForeignKey("manual_task_instance.id", ondelete="CASCADE"),
        nullable=True,
    )
    # The user responsible for the action being logged.  This may be `None`
    # for system-initiated events or for legacy records created before this
    # column existed.
    user_id = Column(String, nullable=True, index=True)
    status = Column(String, nullable=False)
    message = Column(String, nullable=True)
    details = Column(JSONB, nullable=True)

    __table_args__ = (
        Index("ix_manual_task_log_config_id", "config_id"),
        Index("ix_manual_task_log_created_at", "created_at"),
        Index("ix_manual_task_log_instance_id", "instance_id"),
        Index("ix_manual_task_log_status", "status"),
        Index("ix_manual_task_log_task_id", "task_id"),
        Index("ix_manual_task_log_user_id", "user_id"),
    )

    # Relationships
    task = relationship("ManualTask", back_populates="logs", foreign_keys=[task_id])
    config = relationship(
        "ManualTaskConfig",
        back_populates="logs",
        foreign_keys=[config_id],
    )
    instance = relationship("ManualTaskInstance", back_populates="logs")

    @classmethod
    def create_log(
        cls,
        db: Session,
        task_id: str,
        status: "ManualTaskLogStatus",
        message: str,
        user_id: Optional[str] = None,
        config_id: Optional[str] = None,
        instance_id: Optional[str] = None,
        details: Optional[dict[str, Any]] = None,
    ) -> "ManualTaskLog":
        """Create a new task log entry.

        Args:
            db: Database session
            task_id: ID of the task
            status: Status of the log entry
            message: Optional message describing the event
            details: Optional additional details about the event
        """

        data = {
            "task_id": task_id,
            "config_id": config_id,
            "instance_id": instance_id,
            "user_id": user_id or get_user_id(),
            "status": status,
            "message": message,
            "details": details,
        }
        return cls.create(db=db, data=data)

    @classmethod
    def create_error_log(
        cls,
        db: Session,
        task_id: str,
        message: str,
        user_id: Optional[str] = None,
        config_id: Optional[str] = None,
        instance_id: Optional[str] = None,
        details: Optional[dict[str, Any]] = None,
    ) -> "ManualTaskLog":
        """Create a new error log entry.

        Args:
            db: Database session
            task_id: ID of the task
            message: Error message describing what went wrong
            config_id: Optional ID of the configuration
            instance_id: Optional ID of the instance
            details: Optional additional details about the error

        Returns:
            The created error log entry
        """

        return cls.create_log(
            db=db,
            status=ManualTaskLogStatus.error,
            task_id=task_id,
            config_id=config_id,
            instance_id=instance_id,
            user_id=user_id or get_user_id(),
            message=message,
            details=details,
        )



import os
from enum import Enum as EnumType

from loguru import logger

# This is the max file size for downloading the content of an attachment.
# This is an industry standard used by companies like Google and Microsoft.
LARGE_FILE_THRESHOLD = 25 * 1024 * 1024  # 25 MB


class AllowedFileType(EnumType):
    """
    A class that contains the allowed file types and their MIME types.
    """

    pdf = "application/pdf"
    doc = "application/msword"
    docx = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    txt = "text/plain"
    jpg = "image/jpeg"
    jpeg = "image/jpeg"
    png = "image/png"
    xls = "application/vnd.ms-excel"
    xlsx = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
    csv = "text/csv"
    zip = "application/zip"


LOCAL_FIDES_UPLOAD_DIRECTORY = "fides_uploads"


def get_local_filename(file_key: str) -> str:
    """Verifies that the local storage directory exists and returns the local filepath.

    This extra security checks are to prevent directory traversal attacks and "complete business and technical destruction".
    Thanks Claude.

    Args:
        file_key: The key/path for the file

    Returns:
        The full local filepath

    Raises:
        ValueError: If the file_key is invalid or would result in a path outside the upload directory
    """
    # Basic validation
    if not file_key:
        raise ValueError("File key cannot be empty")

    # Security checks before normalization
    if file_key.startswith("/"):
        raise ValueError("Invalid file key: cannot start with '/'")

    # Normalize the path to handle any path separators consistently
    # First normalize using os.path.normpath to handle any redundant separators
    normalized_key = os.path.normpath(file_key)
    # Then convert all separators to forward slashes for consistency
    normalized_key = normalized_key.replace("\\", "/")

    # Additional security: ensure the final path is within the upload directory
    final_path = os.path.join(LOCAL_FIDES_UPLOAD_DIRECTORY, normalized_key)
    if not os.path.abspath(final_path).startswith(
        os.path.abspath(LOCAL_FIDES_UPLOAD_DIRECTORY)
    ):
        raise ValueError(
            "Invalid file key: would result in path outside upload directory"
        )

    # Create all necessary directories
    os.makedirs(os.path.dirname(final_path), exist_ok=True)

    return final_path


def get_allowed_file_type_or_raise(file_key: str) -> str:
    """Verifies that the file type is allowed and returns the MIME type"""
    error_msg = f"Invalid or unallowed file extension: {file_key}"
    if "." not in file_key:
        logger.warning(f"File key {file_key} does not have a file extension")
        raise ValueError(error_msg)
    file_type = file_key.split(".")[-1]
    try:
        return AllowedFileType[file_type].value
    except KeyError:
        raise ValueError(error_msg)



from __future__ import annotations

from typing import Any, Callable, Iterable, List, Optional, Set

from fastapi.applications import FastAPI
from pydantic import SerializeAsAny
from sqlalchemy.orm import Session

from fides.api.custom_types import AnyHttpUrlStringRemovesSlash, URLOriginString
from fides.api.models.application_config import ApplicationConfig
from fides.api.schemas.storage.storage import StorageType
from fides.api.util.cors_middleware_utils import update_cors_middleware
from fides.config import CONFIG


def merge_properties(attribute_names: Iterable[str]) -> Callable:
    """
    Decorator to specify a config proxy class's attributes as config properties
    whose api-set and config-set values should be _merged_ when resolved by the proxy.
    """

    def decorator(cls: ConfigProxyBase) -> ConfigProxyBase:
        cls.merge_properties = set(attribute_names)
        return cls

    return decorator


class ConfigProxyBase:
    """
    Base class that's used to make config proxy classes that correspond
    to our config/settings sub-sections.

    Config proxy classes are a construct to allow for accessing "resolved"
    config properties based on api-set and "traditional" config-set mechanisms.
    Config proxy classes allow these "resolved" properties to be looked up
    as if they were a "normal" pydantic config object, i.e. with dot notation,
    e.g. `ConfigProxy(db).notifications.notification_service_type`

    Merging: to have a given config property's api-set and config-set values to be merged
    on property resolution, use the `@merge_properties` decorator on the config proxy
    class to specify the corresponding attribute name.

    For example, if creating a config proxy class:
    ```
    @merge_properties(["cors_origins"])
    class SecuritySettingsProxy(ConfigProxyBase):
        prefix = "security"
        cors_origins: List[str]
    ```

    and `security.cors_origins` has the following values:
    config-set: `["a", "b"]`
    api-set: `["b", "c"]`

    then, when using the config proxy to access the attribute, it will return
    a merged set of values:
    ```
    >>> ConfigProxy(db).security.cors_origins
    {"a", "b", "c"}
    ```

    """

    prefix: str
    merge_properties: Set[str] = set()

    def __init__(self, db: Session) -> None:
        self._db = db

    def __getattribute__(self, __name: str) -> Any:
        """
        This allows us to retrieve resolved config properties when
        using the config proxy as a "normal" config object, i.e. with dot notation,
        e.g. `config_proxy.notifications.notification_service_type
        """
        if __name in ("_db", "merge_properties", "prefix"):
            return object.__getattribute__(self, __name)
        return ApplicationConfig.get_resolved_config_property(
            self._db,
            f"{self.prefix}.{__name}",
            merge_values=__name in self.merge_properties,
        )


class AdminUISettingsProxy(ConfigProxyBase):
    prefix = "admin_ui"

    enabled: bool
    url: SerializeAsAny[Optional[AnyHttpUrlStringRemovesSlash]] = None


class NotificationSettingsProxy(ConfigProxyBase):
    prefix = "notifications"

    send_request_completion_notification: bool
    send_request_receipt_notification: bool
    send_request_review_notification: bool
    notification_service_type: Optional[str]
    enable_property_specific_messaging: Optional[str]


class ExecutionSettingsProxy(ConfigProxyBase):
    prefix = "execution"

    subject_identity_verification_required: bool
    disable_consent_identity_verification: bool
    require_manual_request_approval: bool

    def __getattribute__(self, name: str) -> Any:
        """
        Overrides base __getattribute__ to provide a dynamic fallback for
        'disable_consent_identity_verification'. The fallback is based on
        'subject_identity_verification_required' only when no explicit value is provided,
        preserving None for unset cases.
        """
        if name == "disable_consent_identity_verification":
            value = super().__getattribute__("disable_consent_identity_verification")
            if value is None:
                subject_verification_required = super().__getattribute__(
                    "subject_identity_verification_required"
                )
                return not subject_verification_required
            return value

        return super().__getattribute__(name)


class StorageSettingsProxy(ConfigProxyBase):
    prefix = "storage"

    active_default_storage_type: StorageType


@merge_properties(["cors_origins"])
class SecuritySettingsProxy(ConfigProxyBase):
    prefix = "security"

    # only valid URLs should be set as cors_origins
    # for advanced usage of non-URLs, e.g. wildcards (`*`), the related
    # `cors_origin_regex` property should be used.
    # this is explicitly _not_ accessible via API - it must be used with care.
    cors_origins: SerializeAsAny[List[URLOriginString]]


class ConsentSettingsProxy(ConfigProxyBase):
    prefix = "consent"

    override_vendor_purposes: bool


class ConfigProxy:
    """
    ConfigProxy instances allow access to "resolved" config properties
    based on resolution between api-set and "traditional" config-set mechanisms.
    ConfigProxy instances allow these "resolved" properties to be looked up
    as if they were a "normal" pydantic config object, i.e. with dot notation,
    e.g. `ConfigProxy(db).notifications.notification_service_type`

    To instantiate a `ConfigProxy`, you must have an active db `Session`,
    since the `ConfigProxy` will rely on the db for property resolution.

    Instantiating a `ConfigProxy` is itself a cheap operation,
    i.e. there is minimal resource overhead with instantiating a new `ConfigProxy`.
    The `ConfigProxy` is a thin wrapper above the given db `Session` -
    it relies on the given db `Session` and for all of its state.

    Lookups (i.e. attribute access) with the `ConfigProxy` do leverage
    the underlying ORM model, but any db calls that are needed should be straightforward -
    it leverages only a fixed single-row table holding the config state.
    """

    def __init__(self, db: Session) -> None:
        self.admin_ui = AdminUISettingsProxy(db)
        self.notifications = NotificationSettingsProxy(db)
        self.execution = ExecutionSettingsProxy(db)
        self.storage = StorageSettingsProxy(db)
        self.security = SecuritySettingsProxy(db)
        self.consent = ConsentSettingsProxy(db)

    def load_current_cors_domains_into_middleware(self, app: FastAPI) -> None:
        """
        Util function that loads the current CORS domains from
        `ConfigProxy` into the  `CORSMiddleware` at runtime.
        """

        # NOTE: `cors_origins` config proxy resolution _merges_ api-set and config-set values, if both present
        current_config_proxy_domains = (
            self.security.cors_origins if self.security.cors_origins is not None else []
        )

        update_cors_middleware(
            app,
            current_config_proxy_domains,
            CONFIG.security.cors_origin_regex,
        )



"""add privacy center config

Revision ID: 2e9aba76c322
Revises: 69e51a460e66
Create Date: 2024-03-14 22:17:45.544448

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "2e9aba76c322"
down_revision = "69e51a460e66"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "plus_privacy_center_config",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("config", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.Column("single_row", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.CheckConstraint(
            "single_row", name="plus_privacy_center_config_single_row_check"
        ),
        sa.UniqueConstraint(
            "single_row", name="plus_privacy_center_config_single_row_unique"
        ),
    )
    op.create_index(
        op.f("ix_plus_privacy_center_config_id"),
        "plus_privacy_center_config",
        ["id"],
        unique=False,
    )


def downgrade():
    op.drop_index(
        op.f("ix_plus_privacy_center_config_id"),
        table_name="plus_privacy_center_config",
    )
    op.drop_table("plus_privacy_center_config")
